import itertools
import logging

from collections import defaultdict
from dataclasses import dataclass
from itertools import combinations
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Sequence
from typing import Tuple

import numpy as np
import pandas as pd

from lightautoml.dataset.roles import CategoryRole
from lightautoml.dataset.roles import ColumnRole
from lightautoml.dataset.roles import NumericRole
from lightautoml.reader.base import RolesDict
from lightautoml.transformers.categorical import categorical_check
from lightautoml.transformers.categorical import encoding_check
from lightautoml.transformers.categorical import multiclass_task_check
from lightautoml.transformers.categorical import oof_task_check
from pyspark.ml import Transformer
from pyspark.ml.feature import OneHotEncoder
from pyspark.sql import Column
from pyspark.sql import Window
from pyspark.sql import functions as sf
from pyspark.sql.types import IntegerType
from sklearn.utils.murmurhash import murmurhash3_32

from sparklightautoml.dataset.roles import NumericVectorOrArrayRole
from sparklightautoml.mlwriters import CommonPickleMLReadable
from sparklightautoml.mlwriters import CommonPickleMLWritable
from sparklightautoml.mlwriters import SparkLabelEncoderTransformerMLReadable
from sparklightautoml.mlwriters import SparkLabelEncoderTransformerMLWritable
from sparklightautoml.transformers.base import SparkBaseEstimator
from sparklightautoml.transformers.base import SparkBaseTransformer
from sparklightautoml.transformers.scala_wrappers.laml_string_indexer import (
    LAMLStringIndexer,
)
from sparklightautoml.transformers.scala_wrappers.laml_string_indexer import (
    LAMLStringIndexerModel,
)
from sparklightautoml.transformers.scala_wrappers.target_encoder_transformer import (
    SparkTargetEncodeTransformer,
)
from sparklightautoml.transformers.scala_wrappers.target_encoder_transformer import (
    TargetEncoderTransformer,
)
from sparklightautoml.utils import SparkDataFrame


logger = logging.getLogger(__name__)


# FIXME SPARK-LAMA: np.nan in str representation is 'nan' while Spark's NaN is 'NaN'. It leads to different hashes.
# FIXME SPARK-LAMA: If udf is defined inside the class, it not works properly.
# "if murmurhash3_32 can be applied to a whole pandas Series, it would be better to make it via pandas_udf"
# https://github.com/fonhorst/LightAutoML/pull/57/files/57c15690d66fbd96f3ee838500de96c4637d59fe#r749534669
murmurhash3_32_udf = sf.udf(
    lambda value: murmurhash3_32(value.replace("NaN", "nan"), seed=42) if value is not None else None,
    IntegerType(),
)


@dataclass
class OOfFeatsMapping:
    folds_column: str
    # count of different categories in the categorical column being processed
    dim_size: int
    # mapping from category to a continious reperesentation found by target encoder
    # category may be represented:
    # - cat (plain category)
    # - dim_size * folds_num + cat
    # mapping: Dict[int, float]
    mapping: np.array


class TypesHelper:
    _ad_hoc_types_mapper = defaultdict(
        lambda: "string",
        {
            "bool": "boolean",
            "int": "int",
            "int8": "int",
            "int16": "int",
            "int32": "int",
            "int64": "int",
            "int128": "bigint",
            "int256": "bigint",
            "integer": "int",
            "uint8": "int",
            "uint16": "int",
            "uint32": "int",
            "uint64": "int",
            "uint128": "bigint",
            "uint256": "bigint",
            "longlong": "long",
            "ulonglong": "long",
            "float16": "float",
            "float": "float",
            "float32": "float",
            "float64": "double",
            "float128": "double",
        },
    )

    _spark_numeric_types_str = ("ShortType", "IntegerType", "LongType", "FloatType", "DoubleType", "DecimalType")


class SparkLabelEncoderEstimator(SparkBaseEstimator, TypesHelper):
    """
    Spark label encoder estimator.
    Returns :class:`~sparklightautoml.transformers.categorical.SparkLabelEncoderTransformer`.
    """

    _fit_checks = (categorical_check,)
    _transform_checks = ()
    _fname_prefix = "le"

    _fillna_val = 0.0

    def __init__(
        self,
        input_cols: Optional[List[str]] = None,
        input_roles: Optional[Dict[str, ColumnRole]] = None,
        subs: Optional[float] = None,
        random_state: Optional[int] = 42,
        do_replace_columns: bool = False,
        output_role: Optional[ColumnRole] = None,
    ):
        if not output_role:
            output_role = CategoryRole(np.int32, label_encoded=True)
        super().__init__(input_cols, input_roles, do_replace_columns=do_replace_columns, output_role=output_role)
        self._input_intermediate_columns = self.getInputCols()
        self._input_intermediate_roles = self.get_input_roles()

    def _fit(self, dataset: SparkDataFrame) -> "SparkLabelEncoderTransformer":
        logger.info(f"[{type(self)} (LE)] fit is started")

        roles = self._input_intermediate_roles
        columns = self._input_intermediate_columns

        # if self._fname_prefix == "inter":
        #     roles = self.get_input_roles()
        #     columns = self.getInputCols()

        indexer = LAMLStringIndexer(
            inputCols=columns,
            outputCols=self.getOutputCols(),
            minFreqs=[roles[col_name].unknown for col_name in columns],
            handleInvalid="keep",
            defaultValue=self._fillna_val,
        )

        self.indexer_model = indexer.fit(dataset)

        logger.info(f"[{type(self)} (LE)] fit is finished")

        return SparkLabelEncoderTransformer(
            input_cols=self.getInputCols(),
            output_cols=self.getOutputCols(),
            input_roles=self.get_input_roles(),
            output_roles=self.get_output_roles(),
            do_replace_columns=self.get_do_replace_columns(),
            indexer_model=self.indexer_model,
        )


class SparkLabelEncoderTransformer(
    SparkBaseTransformer, TypesHelper, SparkLabelEncoderTransformerMLWritable, SparkLabelEncoderTransformerMLReadable
):
    """
    Simple Spark version of `LabelEncoder`.

    Labels are integers from 1 to n.
    """

    _transform_checks = ()
    _fname_prefix = "le"

    _fillna_val = 0.0

    def __init__(
        self,
        input_cols: List[str],
        output_cols: List[str],
        input_roles: RolesDict,
        output_roles: RolesDict,
        do_replace_columns: bool,
        indexer_model: LAMLStringIndexerModel,
    ):
        super().__init__(input_cols, output_cols, input_roles, output_roles, do_replace_columns)
        self.indexer_model = indexer_model
        self._input_intermediate_columns = self.getInputCols()
        self._input_intermediate_roles = self.get_input_roles()

    def _transform(self, dataset: SparkDataFrame) -> SparkDataFrame:
        logger.info(f"[{type(self)} (LE)] transform is started")

        columns = self._input_intermediate_columns
        out_columns = self.getOutputCols()

        model: LAMLStringIndexerModel = (
            self.indexer_model.setDefaultValue(self._fillna_val)
            .setHandleInvalid("keep")
            .setInputCols(columns)
            .setOutputCols(out_columns)
        )

        logger.info(f"[{type(self)} (LE)] Transform is finished")

        output = model.transform(dataset)

        rest_cols = [col for col in output.columns if col not in self.getOutputCols()]
        out_cols = [sf.col(col).astype(IntegerType()).alias(col) for col in self.getOutputCols()]

        output = output.select([*rest_cols, *out_cols])

        return output


class SparkOrdinalEncoderEstimator(SparkLabelEncoderEstimator):
    """
    Spark ordinal encoder estimator.
    Returns :class:`~sparklightautoml.transformers.categorical.SparkOrdinalEncoderTransformer`.
    """

    _fit_checks = (categorical_check,)
    _fname_prefix = "ord"
    _fillna_val = float("nan")

    def __init__(
        self,
        input_cols: Optional[List[str]] = None,
        input_roles: Optional[Dict[str, ColumnRole]] = None,
        subs: Optional[float] = None,
        random_state: Optional[int] = 42,
    ):
        super().__init__(input_cols, input_roles, subs, random_state, output_role=NumericRole(np.float32))
        self.dicts = None
        self._use_cols = self.getInputCols()

    def _fit(self, dataset: SparkDataFrame) -> "Transformer":
        logger.info(f"[{type(self)} (ORD)] fit is started")

        cols_to_process = [
            col for col in self.getInputCols() if str(dataset.schema[col].dataType) not in self._spark_numeric_types_str
        ]

        min_freqs = [self._input_intermediate_roles[col].unknown for col in cols_to_process]

        indexer = LAMLStringIndexer(
            stringOrderType="alphabetAsc",
            inputCols=cols_to_process,
            outputCols=[f"{self._fname_prefix}__{col}" for col in cols_to_process],
            minFreqs=min_freqs,
            handleInvalid="keep",
            defaultValue=self._fillna_val,
            nanLast=True,  # Only for ORD
        )

        self.indexer_model = indexer.fit(dataset)

        logger.info(f"[{type(self)} (ORD)] fit is finished")

        return SparkOrdinalEncoderTransformer(
            input_cols=self.getInputCols(),
            output_cols=self.getOutputCols(),
            input_roles=self.get_input_roles(),
            output_roles=self.get_output_roles(),
            do_replace_columns=self.get_do_replace_columns(),
            indexer_model=self.indexer_model,
        )


class SparkOrdinalEncoderTransformer(SparkLabelEncoderTransformer):
    """
    Spark version of :class:`~lightautoml.transformers.categorical.OrdinalEncoder`.

    Encoding ordinal categories into numbers.
    Number type categories passed as is,
    object type sorted in ascending lexicographical order.
    """

    _transform_checks = ()
    _fname_prefix = "ord"
    _fillna_val = float("nan")

    def __init__(
        self,
        input_cols: List[str],
        output_cols: List[str],
        input_roles: RolesDict,
        output_roles: RolesDict,
        do_replace_columns: bool,
        indexer_model: LAMLStringIndexerModel,
    ):
        super().__init__(input_cols, output_cols, input_roles, output_roles, do_replace_columns, indexer_model)

    def _transform(self, dataset: SparkDataFrame) -> SparkDataFrame:
        logger.info(f"[{type(self)} (ORD)] transform is started")

        input_columns = self.getInputCols()
        output_columns = self.getOutputCols()

        cols_to_process = [
            col for col in self.getInputCols() if str(dataset.schema[col].dataType) not in self._spark_numeric_types_str
        ]

        # TODO: Fix naming
        transformed_cols = [f"{self._fname_prefix}__{col}" for col in cols_to_process]

        model: LAMLStringIndexerModel = (
            self.indexer_model.setDefaultValue(self._fillna_val)
            .setHandleInvalid("keep")
            .setInputCols(cols_to_process)
            .setOutputCols(transformed_cols)
        )

        indexed_dataset = model.transform(dataset)

        for input_col, output_col in zip(input_columns, output_columns):
            if output_col in transformed_cols:
                continue
            indexed_dataset = indexed_dataset.withColumn(output_col, sf.col(input_col))

        indexed_dataset = indexed_dataset.replace(float("nan"), 0.0, subset=output_columns)

        logger.info(f"[{type(self)} (ORD)] Transform is finished")

        # output = self._make_output_df(indexed_dataset, self.getOutputCols())
        output = indexed_dataset

        return output


class SparkFreqEncoderEstimator(SparkLabelEncoderEstimator):
    """
    Calculates frequency in train data and
    produces :class:`~sparklightautoml.transformers.categorical.SparkFreqEncoderTransformer` instance.
    """

    _fit_checks = (categorical_check,)
    _transform_checks = ()
    _fname_prefix = "freq"

    _fillna_val = 1

    def __init__(self, input_cols: List[str], input_roles: RolesDict, do_replace_columns: bool = False):
        super().__init__(input_cols, input_roles, do_replace_columns, output_role=NumericRole(np.float32))

    def _fit(self, dataset: SparkDataFrame) -> "SparkFreqEncoderTransformer":
        logger.info(f"[{type(self)} (FE)] fit is started")

        indexer = LAMLStringIndexer(
            inputCols=self._input_intermediate_columns,
            outputCols=self.getOutputCols(),
            minFreqs=[1 for _ in self._input_intermediate_columns],
            handleInvalid="keep",
            defaultValue=self._fillna_val,
            freqLabel=True,  # Only for FREQ encoder
        )

        self.indexer_model = indexer.fit(dataset)

        logger.info(f"[{type(self)} (FE)] fit is finished")

        return SparkFreqEncoderTransformer(
            input_cols=self.getInputCols(),
            output_cols=self.getOutputCols(),
            input_roles=self.get_input_roles(),
            output_roles=self.get_output_roles(),
            do_replace_columns=self.get_do_replace_columns(),
            indexer_model=self.indexer_model,
        )


class SparkFreqEncoderTransformer(SparkLabelEncoderTransformer):
    """
    Labels are encoded with frequency in train data.

    Labels are integers from 1 to n.
    """

    _fit_checks = (categorical_check,)
    _transform_checks = ()
    _fname_prefix = "freq"
    _fillna_val = 1.0

    def __init__(
        self,
        input_cols: List[str],
        output_cols: List[str],
        input_roles: RolesDict,
        output_roles: RolesDict,
        do_replace_columns: bool,
        indexer_model: LAMLStringIndexerModel,
    ):
        super().__init__(input_cols, output_cols, input_roles, output_roles, do_replace_columns, indexer_model)


class SparkCatIntersectionsHelper:
    """Helper class for :class:`~sparklightautoml.transformers.categorical.SparkCatIntersectionsEstimator` and
    :class:`~sparklightautoml.transformers.categorical.SparkCatIntersectionsTransformer`.
    """

    _fname_prefix = "inter"

    # noinspection PyMethodMayBeStatic
    def _make_col_name(self, cols: Sequence[str]) -> str:
        return f"({'__'.join(cols)})"

    def _make_category(self, cols: Sequence[str]) -> Column:
        lit = sf.lit("_")
        col_name = self._make_col_name(cols)
        columns_for_concat = []
        for col in cols:
            columns_for_concat.append(sf.col(col))
            columns_for_concat.append(lit)
        columns_for_concat = columns_for_concat[:-1]

        # return murmurhash3_32_udf(sf.concat(*columns_for_concat)).alias(col_name)
        return sf.hash(sf.concat(*columns_for_concat)).alias(col_name)

    def _build_df(
        self, df: SparkDataFrame, intersections: Optional[Sequence[Sequence[str]]]
    ) -> Tuple[SparkDataFrame, List[str]]:
        col_names = [self._make_col_name(comb) for comb in intersections]
        columns_to_select = [
            self._make_category(comb).alias(col_name) for comb, col_name in zip(intersections, col_names)
        ]
        df = df.select("*", *columns_to_select)
        return df, col_names


class SparkCatIntersectionsEstimator(SparkCatIntersectionsHelper, SparkLabelEncoderEstimator):
    """
    Combines categorical features
    and fits :class:`~sparklightautoml.transformers.categorical.SparkLabelEncoderEstimator`.
    Returns :class:`~sparklightautoml.transformers.categorical.SparkCatIntersectionsTransformer`.
    """

    _fit_checks = (categorical_check,)
    _transform_checks = ()
    _fname_prefix = "inter"

    def __init__(
        self,
        input_cols: List[str],
        input_roles: Dict[str, ColumnRole],
        intersections: Optional[Sequence[Sequence[str]]] = None,
        max_depth: int = 2,
        do_replace_columns: bool = False,
    ):
        super().__init__(
            input_cols,
            input_roles,
            do_replace_columns=do_replace_columns,
            output_role=CategoryRole(np.int32, label_encoded=True),
        )
        self.intersections = intersections
        self.max_depth = max_depth

        if self.intersections is None:
            self.intersections = []
            for i in range(2, min(self.max_depth, len(self.getInputCols())) + 1):
                self.intersections.extend(list(combinations(self.getInputCols(), i)))

        self._input_roles = {
            f"{self._make_col_name(comb)}": CategoryRole(
                np.int32,
                unknown=max((self.get_input_roles()[x].unknown for x in comb)),
                label_encoded=True,
            )
            for comb in self.intersections
        }
        self._input_columns = sorted(list(self._input_roles.keys()))

        out_roles = {f"{self._fname_prefix}__{f}": role for f, role in self._input_roles.items()}

        self.set(self.outputCols, list(out_roles.keys()))
        self.set(self.outputRoles, out_roles)

    def _fit(self, df: SparkDataFrame) -> Transformer:
        logger.info(f"[{type(self)} (CI)] fit is started")
        logger.debug(f"Calculating (CI) for input columns: {self.getInputCols()}")

        inter_df, inter_cols = self._build_df(df, self.intersections)

        self._input_intermediate_roles = {
            col: self.get_input_roles()[elts[0]] for col, elts in zip(inter_cols, self.intersections)
        }
        self._input_intermediate_columns = inter_cols

        super()._fit(inter_df)

        logger.info(f"[{type(self)} (CI)] fit is finished")

        return SparkCatIntersectionsTransformer(
            input_cols=self.getInputCols(),
            output_cols=self.getOutputCols(),
            input_roles=self.get_input_roles(),
            output_roles=self.get_output_roles(),
            do_replace_columns=self.get_do_replace_columns(),
            indexer_model=self.indexer_model,
            intersections=self.intersections,
        )


class SparkCatIntersectionsTransformer(SparkCatIntersectionsHelper, SparkLabelEncoderTransformer):
    """
    Combines category columns and encode with label encoder.
    """

    _fit_checks = (categorical_check,)
    _transform_checks = ()

    _fillna_val = 0

    def __init__(
        self,
        input_cols: List[str],
        output_cols: List[str],
        input_roles: RolesDict,
        output_roles: RolesDict,
        do_replace_columns: bool,
        indexer_model: LAMLStringIndexerModel,
        intersections: Optional[Sequence[Sequence[str]]] = None,
    ):
        super().__init__(input_cols, output_cols, input_roles, output_roles, do_replace_columns, indexer_model)
        self.intersections = intersections

    def _transform(self, df: SparkDataFrame) -> SparkDataFrame:
        inter_df, self._input_intermediate_columns = self._build_df(df, self.intersections)

        out_df = super()._transform(inter_df)

        out_df = out_df.drop(*self._input_intermediate_columns)
        return out_df


class SparkOHEEncoderEstimator(SparkBaseEstimator):
    """
    Simple OneHotEncoder over label encoded categories.
    """

    _fit_checks = (categorical_check, encoding_check)
    _transform_checks = ()
    _fname_prefix = "ohe"

    @property
    def features(self) -> List[str]:
        """Features list."""
        return self.getInputCols()

    def __init__(
        self,
        input_cols: List[str],
        input_roles: Dict[str, ColumnRole],
        do_replace_columns: bool = False,
        make_sparse: Optional[bool] = None,
        total_feats_cnt: Optional[int] = None,
        dtype: type = np.float32,
    ):
        """
        Args:
            make_sparse: Create sparse matrix.
            total_feats_cnt: Initial features number.
            dtype: Dtype of new features.
        """
        super().__init__(input_cols, input_roles, do_replace_columns=do_replace_columns, output_role=None)

        self._make_sparse = make_sparse
        self._total_feats_cnt = total_feats_cnt
        self.dtype = dtype

        if self._make_sparse is None:
            assert self._total_feats_cnt is not None, "Param total_feats_cnt should be defined if make_sparse is None"

        self._ohe_transformer_and_roles: Optional[Tuple[Transformer, Dict[str, ColumnRole]]] = None

    def _fit(self, sdf: SparkDataFrame) -> Transformer:
        """Calc output shapes.
        Automatically do ohe in sparse form if approximate fill_rate < `0.2`.
        Args:
            sdf: Spark dataframe of categorical features.
        Returns:
            self.
        """

        maxs = [sf.max(c).alias(f"max_{c}") for c in self.getInputCols()]
        mins = [sf.min(c).alias(f"min_{c}") for c in self.getInputCols()]
        mm = sdf.select(maxs + mins).first().asDict()

        ohe = OneHotEncoder(inputCols=self.getInputCols(), outputCols=self.getOutputCols(), handleInvalid="error")
        transformer = ohe.fit(sdf)

        roles = {
            f"{self._fname_prefix}__{c}": NumericVectorOrArrayRole(
                size=mm[f"max_{c}"] - mm[f"min_{c}"] + 1,
                element_col_name_template=[
                    f"{self._fname_prefix}_{i}__{c}" for i in np.arange(mm[f"min_{c}"], mm[f"max_{c}"] + 1)
                ],
            )
            for c in self.getInputCols()
        }

        self._ohe_transformer_and_roles = (transformer, roles)

        return OHEEncoderTransformer(
            transformer,
            input_cols=self.getInputCols(),
            output_cols=self.getOutputCols(),
            input_roles=self.get_input_roles(),
            output_roles=roles,
        )


class OHEEncoderTransformer(SparkBaseTransformer, CommonPickleMLWritable, CommonPickleMLReadable):
    """OHEEncoder Transformer"""

    _fit_checks = (categorical_check, encoding_check)
    _transform_checks = ()
    _fname_prefix = "ohe"

    @property
    def features(self) -> List[str]:
        """Features list."""
        return self.getInputCols()

    def __init__(
        self,
        ohe_transformer: Transformer,
        input_cols: List[str],
        output_cols: List[str],
        input_roles: RolesDict,
        output_roles: RolesDict,
        do_replace_columns: bool = False,
    ):
        super().__init__(input_cols, output_cols, input_roles, output_roles, do_replace_columns)
        self._ohe_transformer = ohe_transformer

    def _transform(self, sdf: SparkDataFrame) -> SparkDataFrame:
        """Transform categorical dataset to ohe.
        Args:
            sdf: Spark dataframe of categorical features.
        Returns:
            Numpy dataset with encoded labels.
        """
        output = self._ohe_transformer.transform(sdf)

        return output


def te_mapping_udf(broadcasted_dict):
    def f(folds, current_column):
        values_dict = broadcasted_dict.value
        try:
            return values_dict[f"{folds}_{current_column}"]
        except KeyError:
            return np.nan

    return sf.udf(f, "double")


class SparkTargetEncoderEstimator(SparkBaseEstimator):
    """
    Spark target encoder estimator.
    Returns :class:`~sparklightautoml.transformers.categorical.SparkTargetEncoderTransformer`.
    """

    _fit_checks = (categorical_check, oof_task_check, encoding_check)
    _transform_checks = ()
    _fname_prefix = "oof"

    def __init__(
        self,
        input_cols: List[str],
        input_roles: Dict[str, ColumnRole],
        alphas: Sequence[float] = (0.5, 1.0, 2.0, 5.0, 10.0, 50.0, 250.0, 1000.0),
        task_name: Optional[str] = None,
        folds_column: Optional[str] = None,
        target_column: Optional[str] = None,
        do_replace_columns: bool = False,
    ):
        super().__init__(
            input_cols, input_roles, do_replace_columns, NumericRole(np.float32, prob=task_name == "binary")
        )
        self.alphas = alphas
        self._task_name = task_name
        self._folds_column = folds_column
        self._target_column = target_column

    @staticmethod
    def score_func_binary(target, candidate) -> float:
        return -(target * np.log(candidate) + (1 - target) * np.log(1 - candidate))

    @staticmethod
    def score_func_reg(target, candidate) -> float:
        return (target - candidate) ** 2

    def _fit(self, dataset: SparkDataFrame) -> "SparkBaseTransformer":
        logger.info(f"[{type(self)} (TE)] fit_transform is started")

        assert self._target_column in dataset.columns, "Target should be presented in the dataframe"
        assert self._folds_column in dataset.columns, "Folds should be presented in the dataframe"

        self.encodings: Dict[str, np.ndarray] = dict()
        oof_feats_encoding: Dict[str, OOfFeatsMapping] = dict()

        sdf = dataset

        _fc = sf.col(self._folds_column)
        _tc = sf.col(self._target_column)

        logger.debug("Calculating totals (TE)")
        n_folds, prior, total_target_sum, total_count = sdf.select(
            sf.max(_fc) + 1, sf.mean(_tc.cast("double")), sf.sum(_tc).cast("double"), sf.count(_tc)
        ).first()

        logger.debug("Calculating folds priors (TE)")
        folds_prior_pdf = (
            sdf.groupBy(_fc)
            .agg(((total_target_sum - sf.sum(_tc)) / (total_count - sf.count(_tc))).alias("_folds_prior"))
            .collect()
        )

        def binary_score(col_name: str):
            return sf.mean(-(_tc * sf.log(col_name) + (1 - _tc) * sf.log(1 - sf.col(col_name)))).alias(col_name)

        def reg_score(col_name: str):
            return sf.mean(sf.pow((_tc - sf.col(col_name)), sf.lit(2))).alias(col_name)

        logger.debug("Starting processing features")
        feature_count = len(self.getInputCols())
        for i, feature in enumerate(self.getInputCols()):
            logger.debug(f"Processing feature {feature}({i}/{feature_count})")

            _cur_col = sf.col(feature)
            (dim_size,) = sdf.select((sf.max(_cur_col) + 1).astype("int").alias("dim_size")).first()

            logger.debug(f"Dim size of feature {feature}: {dim_size}")

            window_spec = Window.partitionBy(_cur_col)
            f_df = sdf.groupBy(_cur_col, _fc).agg(sf.sum(_tc).alias("f_sum"), sf.count(_tc).alias("f_count")).cache()

            oof_df = f_df.select(
                _cur_col,
                _fc,
                (sf.sum("f_sum").over(window_spec) - sf.col("f_sum")).alias("oof_sum"),
                (sf.sum("f_count").over(window_spec) - sf.col("f_count")).alias("oof_count"),
            )

            logger.debug(f"Creating maps column for fold priors (size={len(folds_prior_pdf)}) (TE)")
            mapping = {row[self._folds_column]: row["_folds_prior"] for row in folds_prior_pdf}
            folds_prior_exp = sf.create_map(*[sf.lit(x) for x in itertools.chain(*mapping.items())])

            logger.debug(f"Creating candidate columns (count={len(self.alphas)}) (TE)")
            candidates_cols = [
                ((sf.col("oof_sum") + sf.lit(alpha) * folds_prior_exp[_fc]) / (sf.col("oof_count") + sf.lit(alpha)))
                .cast("double")
                .alias(f"candidate_{i}")
                for i, alpha in enumerate(self.alphas)
            ]

            candidates_df = oof_df.select(_cur_col, _fc, *candidates_cols).cache()

            score_func = binary_score if self._task_name == "binary" else reg_score

            logger.debug("Calculating scores (TE)")
            scores = (
                sdf.join(candidates_df, on=[feature, self._folds_column])
                .select(*[score_func(f"candidate_{i}") for i, alpha in enumerate(self.alphas)])
                .first()
                .asDict()
            )
            logger.debug(f"Scores have been calculated (size={len(scores)}) (TE)")

            seq_scores = [scores[f"candidate_{i}"] for i, alpha in enumerate(self.alphas)]
            best_alpha_idx = np.argmin(seq_scores)
            best_alpha = self.alphas[best_alpha_idx]

            logger.debug("Collecting encodings (TE)")
            encoding_df = f_df.groupby(_cur_col).agg(
                ((sf.sum("f_sum") + best_alpha * prior) / (sf.sum("f_count") + best_alpha)).alias("encoding")
            )
            encoding = encoding_df.toPandas()
            logger.debug(f"Encodings have been collected (size={len(encoding)}) (TE)")
            f_df.unpersist()

            mapping = np.zeros(dim_size, dtype=np.float64)
            np.add.at(mapping, encoding[feature].astype(np.int32).to_numpy(), encoding["encoding"])
            self.encodings[feature] = mapping

            logger.debug("Collecting oof_feats (TE)")
            oof_feats_df = candidates_df.select(_cur_col, _fc, sf.col(f"candidate_{best_alpha_idx}").alias("encoding"))
            oof_feats = oof_feats_df.toPandas()
            logger.debug(f"oof_feats have been collected (size={len(oof_feats)}) (TE)")

            candidates_df.unpersist()

            mapping = np.zeros((n_folds, dim_size), dtype=np.float64)
            np.add.at(
                mapping,
                (
                    oof_feats[self._folds_column].astype(np.int32).to_numpy(),
                    oof_feats[feature].astype(np.int32).to_numpy(),
                ),
                oof_feats["encoding"],
            )

            oof_feats = OOfFeatsMapping(folds_column=self._folds_column, dim_size=dim_size, mapping=mapping)

            oof_feats_encoding[feature] = oof_feats

            logger.debug(f"[{type(self)} (TE)] Encodings have been calculated")

        logger.info(f"[{type(self)} (TE)] fit_transform is finished")

        return SparkTargetEncodeTransformer(
            tet=TargetEncoderTransformer.create(
                enc={col: mapping.tolist() for col, mapping in self.encodings.items()},
                oof_enc={col: mapping.mapping.tolist() for col, mapping in oof_feats_encoding.items()},
                fold_column=self._folds_column,
                apply_oof=True,
                input_cols=list(self.get_input_roles().keys()),
                output_cols=list(self.get_output_roles().keys()),
            ),
            input_roles=self.get_input_roles(),
            output_roles=self.get_output_roles(),
        )


def mcte_transform_udf(broadcasted_dict):
    def f(target, current_column):
        values_dict = broadcasted_dict.value
        try:
            return values_dict[(target, current_column)]
        except KeyError:
            return np.nan

    return sf.udf(f, "double")


class SparkMulticlassTargetEncoderEstimator(SparkBaseEstimator):
    """
    Spark multiclass target encoder estimator.
    Returns :class:`~sparklightautoml.transformers.categorical.SparkMultiTargetEncoderTransformer`.
    """

    _fit_checks = (categorical_check, multiclass_task_check, encoding_check)
    _transform_checks = ()
    _fname_prefix = "multioof"

    def __init__(
        self,
        input_cols: List[str],
        input_roles: Dict[str, ColumnRole],
        alphas: Sequence[float] = (0.5, 1.0, 2.0, 5.0, 10.0, 50.0, 250.0, 1000.0),
        task_name: Optional[str] = None,
        folds_column: Optional[str] = None,
        target_column: Optional[str] = None,
        do_replace_columns: bool = False,
    ):
        super().__init__(input_cols, input_roles, do_replace_columns, NumericRole(np.float32, prob=True))
        self.alphas = alphas
        self._task_name = task_name
        self._folds_column = folds_column
        self._target_column = target_column

    def _fit(self, dataset: SparkDataFrame) -> "SparkMultiTargetEncoderTransformer":
        logger.info(f"[{type(self)} (MCTE)] fit_transform is started")

        assert self._target_column in dataset.columns, "Target should be presented in the dataframe"
        assert self._folds_column in dataset.columns, "Folds should be presented in the dataframe"

        self.encodings = []

        df = dataset

        _fc = sf.col(self._folds_column)
        _tc = sf.col(self._target_column)

        tcn = self._target_column
        fcn = self._folds_column

        agg = df.groupBy([_fc, _tc]).count().toPandas().sort_values(by=[fcn, tcn])

        rows_count = agg["count"].sum()
        prior = agg.groupby(tcn).agg({"count": sum})

        prior["prior"] = prior["count"] / float(rows_count)
        prior = prior.to_dict()["prior"]

        agg["tt_sum"] = agg[tcn].map(agg[[tcn, "count"]].groupby(tcn).sum()["count"].to_dict()) - agg["count"]
        agg["tf_sum"] = rows_count - agg[fcn].map(agg[[fcn, "count"]].groupby(fcn).sum()["count"].to_dict())

        agg["folds_prior"] = agg["tt_sum"] / agg["tf_sum"]
        folds_prior_dict = agg[[fcn, tcn, "folds_prior"]].groupby([fcn, tcn]).max().to_dict()["folds_prior"]

        # Folds column unique values
        fcvs = sorted(list(set([fold for fold, target in folds_prior_dict.keys()])))
        # Target column unique values
        tcvs = sorted(list(set([target for fold, target in folds_prior_dict.keys()])))

        # cols_to_select = []

        for ccn in self.getInputCols():
            logger.debug(f"[{type(self)} (MCTE)] column {ccn}")

            _cc = sf.col(ccn)

            col_agg = df.groupby(_fc, _tc, _cc).count().toPandas()
            col_agg_dict = col_agg.groupby([ccn, fcn, tcn]).sum().to_dict()["count"]
            t_sum_dict = col_agg[[ccn, tcn, "count"]].groupby([ccn, tcn]).sum().to_dict()["count"]
            f_count_dict = col_agg[[ccn, fcn, "count"]].groupby([ccn, fcn]).sum().to_dict()["count"]
            t_count_dict = col_agg[[ccn, "count"]].groupby([ccn]).sum().to_dict()["count"]

            alphas_values = dict()
            # Current column unique values
            ccvs = sorted(col_agg[ccn].unique())

            for column_value in ccvs:
                for fold in fcvs:
                    oof_count = t_count_dict.get(column_value, 0) - f_count_dict.get((column_value, fold), 0)
                    for target in tcvs:
                        oof_sum = t_sum_dict.get((column_value, target), 0) - col_agg_dict.get(
                            (column_value, fold, target), 0
                        )
                        alphas_values[(column_value, fold, target)] = [
                            (oof_sum + a * folds_prior_dict[(fold, target)]) / (oof_count + a) for a in self.alphas
                        ]

            def make_candidates(x):
                fold, target, col_val, count = x
                values = alphas_values[(col_val, fold, target)]
                for i, a in enumerate(self.alphas):
                    x[f"alpha_{i}"] = values[i]
                return x

            candidates_df = col_agg.apply(make_candidates, axis=1)

            best_alpha_index = np.array(
                [
                    (-np.log(candidates_df[f"alpha_{i}"]) * candidates_df["count"]).sum()
                    for i, a in enumerate(self.alphas)
                ]
            ).argmin()

            column_encodings_dict = (
                pd.DataFrame(
                    [
                        [
                            ccv,
                            tcv,
                            (t_sum_dict.get((ccv, tcv), 0) + self.alphas[best_alpha_index] * prior[tcv])
                            / (t_count_dict[ccv] + self.alphas[best_alpha_index]),
                        ]
                        for (ccv, fcv, tcv), _ in alphas_values.items()
                    ],
                    columns=[ccn, tcn, "encoding"],
                )
                .groupby([tcn, ccn])
                .max()
                .to_dict()["encoding"]
            )

            self.encodings.append(column_encodings_dict)

        logger.info(f"[{type(self)} (MCTE)] fit_transform is finished")

        return SparkMultiTargetEncoderTransformer(
            encodings=self.encodings,
            input_cols=self.getInputCols(),
            input_roles=self.get_input_roles(),
            output_cols=self.getOutputCols(),
            output_roles=self.get_output_roles(),
            do_replace_columns=self.get_do_replace_columns(),
        )


class SparkMultiTargetEncoderTransformer(SparkBaseTransformer, CommonPickleMLWritable, CommonPickleMLReadable):
    """
    Spark multiclass target encoder transformer.
    """

    _fit_checks = (categorical_check, multiclass_task_check, encoding_check)
    _transform_checks = ()
    _fname_prefix = "multioof"

    def __init__(
        self,
        encodings: List[Dict[str, Any]],
        input_cols: List[str],
        output_cols: List[str],
        input_roles: RolesDict,
        output_roles: RolesDict,
        do_replace_columns: bool = False,
    ):
        super().__init__(input_cols, output_cols, input_roles, output_roles, do_replace_columns)
        self._encodings = encodings

    def _transform(self, dataset: SparkDataFrame) -> SparkDataFrame:
        cols_to_select = []
        logger.info(f"[{type(self)} (MCTE)] transform is started")

        sc = dataset.sql_ctx.sparkSession.sparkContext

        for i, (col_name, out_name) in enumerate(zip(self.getInputCols(), self.getOutputCols())):
            _cc = sf.col(col_name)
            logger.debug(f"[{type(self)} (MCTE)] transform map size for column {col_name}: {len(self._encodings[i])}")

            enc = self._encodings[i]
            values = sc.broadcast(enc)
            for tcv in {tcv for tcv, _ in enc.keys()}:
                cols_to_select.append(mcte_transform_udf(values)(sf.lit(tcv), _cc).alias(out_name))

        output = self._make_output_df(dataset, cols_to_select)

        logger.info(f"[{type(self)} (MCTE)] transform is finished")

        return output



import logging

from abc import ABC
from abc import abstractmethod
from contextlib import contextmanager
from copy import deepcopy
from dataclasses import dataclass
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import TypeVar
from typing import Union

from sparklightautoml.computations.utils import deecopy_tviter_without_dataset
from sparklightautoml.dataset.base import SparkDataset
from sparklightautoml.validation.base import SparkBaseTrainValidIterator


logger = logging.getLogger(__name__)

T = TypeVar("T")
R = TypeVar("R")


@dataclass
class ComputationSlot:
    id: str
    dataset: Optional[SparkDataset] = None
    num_tasks: Optional[int] = None
    num_threads_per_executor: Optional[int] = None


class ComputationsSession(ABC):
    @abstractmethod
    def __enter__(self):
        ...

    @abstractmethod
    def __exit__(self, exc_type, exc_val, exc_tb):
        ...

    @abstractmethod
    @contextmanager
    def allocate(self) -> ComputationSlot:
        """
        Thread safe method
        Returns:

        """
        ...

    def map_and_compute(self, func: Callable[[R], T], tasks: List[R]) -> List[T]:
        ...

    def compute(self, tasks: List[Callable[[], T]]) -> List[T]:
        ...


class ComputationsManager(ABC):
    @property
    @abstractmethod
    def parallelism(self) -> int:
        ...

    @contextmanager
    @abstractmethod
    def session(self, dataset: Optional[SparkDataset] = None) -> ComputationsSession:
        ...

    def compute_on_dataset(self, dataset: SparkDataset, tasks: List[Callable[[ComputationSlot], T]]) -> List[T]:
        with self.session(dataset) as session:

            def _task_wrap(task):
                with session.allocate() as slot:
                    return task(slot)

            return session.map_and_compute(_task_wrap, tasks)

    def compute(self, tasks: List[Callable[[], T]]) -> List[T]:
        # use session here for synchronization purpose
        with self.session() as session:
            return session.compute(tasks)

    def compute_folds(
        self, train_val_iter: SparkBaseTrainValidIterator, task: Callable[[int, ComputationSlot], T]
    ) -> List[T]:
        tv_iter = deecopy_tviter_without_dataset(train_val_iter)

        with self.session(train_val_iter.train) as session:

            def _task_wrap(fold_id: int):
                with session.allocate() as slot:
                    local_tv_iter = deepcopy(tv_iter)
                    local_tv_iter.train = slot.dataset
                    slot.dataset = None
                    new_slot = deepcopy(slot)
                    new_slot.dataset = local_tv_iter[fold_id]
                    slot.dataset = local_tv_iter.train
                    return task(fold_id, new_slot)

            fold_ids = list(range(len(train_val_iter)))
            return session.map_and_compute(_task_wrap, fold_ids)


# either parallelism settings or manager
ComputationsSettings = Union[Dict[str, Any], ComputationsManager]



import logging

from abc import ABC
from copy import copy
from typing import List
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import cast

import numpy as np

from lightautoml.automl.blend import WeightedBlender
from lightautoml.dataset.roles import ColumnRole
from lightautoml.dataset.roles import NumericRole
from lightautoml.reader.base import RolesDict
from pyspark.ml import Pipeline
from pyspark.ml import PipelineModel
from pyspark.ml import Transformer
from pyspark.ml.feature import SQLTransformer

from sparklightautoml.dataset.base import SparkDataset
from sparklightautoml.dataset.roles import NumericVectorOrArrayRole
from sparklightautoml.ml_algo.base import AveragingTransformer
from sparklightautoml.pipelines.base import TransformerInputOutputRoles
from sparklightautoml.pipelines.ml.base import SparkMLPipeline
from sparklightautoml.tasks.base import DEFAULT_PREDICTION_COL_NAME
from sparklightautoml.tasks.base import SparkTask
from sparklightautoml.transformers.base import DropColumnsTransformer
from sparklightautoml.utils import ColumnsSelectorTransformer


logger = logging.getLogger(__name__)


class SparkBlender(TransformerInputOutputRoles, ABC):
    """Basic class for blending.

    Blender learns how to make blend
    on sequence of prediction datasets and prune pipes,
    that are not used in final blend.

    """

    def __init__(self):
        self._transformer = None
        self._single_prediction_col_name = DEFAULT_PREDICTION_COL_NAME
        self._pred_role: Optional[ColumnRole] = None
        self._input_roles: Optional[RolesDict] = None
        self._output_roles: Optional[RolesDict] = None
        self._task: Optional[SparkTask] = None
        self._service_columns: Optional[List[str]] = None
        super().__init__()

    @property
    def input_roles(self) -> Optional[RolesDict]:
        """Returns dict of input roles"""
        return self._input_roles

    @property
    def output_roles(self) -> Optional[RolesDict]:
        """Returns dict of output roles"""
        return self._output_roles

    def _get_service_columns(self) -> List[str]:
        return self._service_columns

    def transformer(self, *args, **kwargs) -> Optional[Transformer]:
        """Returns Spark MLlib Transformer.
        Represents a Transformer with fitted models."""

        assert self._transformer is not None, "Pipeline is not fitted!"

        return self._transformer

    def _build_transformer(self, *args, **kwargs) -> Optional[Transformer]:
        return self._transformer

    def fit_predict(
        self, predictions: SparkDataset, pipes: Sequence[SparkMLPipeline]
    ) -> Tuple[SparkDataset, Sequence[SparkMLPipeline]]:
        logger.info(f"Blender {type(self)} starting fit_predict")

        self._set_metadata(predictions, pipes)

        if len(pipes) == 1 and len(pipes[0].ml_algos) == 1:
            statement = (
                f"SELECT *, {pipes[0].ml_algos[0].prediction_feature} "
                f"AS {self._single_prediction_col_name} FROM __THIS__"
            )

            logger.info(f"Select prediction columns with query: {statement}")

            self._transformer = Pipeline(
                stages=[
                    SQLTransformer(statement=statement),
                    DropColumnsTransformer(remove_cols=[pipes[0].ml_algos[0].prediction_feature]),
                ]
            ).fit(predictions.data)

            preds = predictions.empty()
            preds.set_data(
                self._transformer.transform(predictions.data),
                list(self.output_roles.keys()),
                self.output_roles,
                name=type(self).__name__,
            )

            return preds, pipes

        result = self._fit_predict(predictions, pipes)

        self._output_roles = result[0].roles
        self._service_columns = predictions.service_columns

        logger.info(f"Blender {type(self)} finished fit_predict")

        return result

    def predict(self, predictions: SparkDataset) -> SparkDataset:
        return self._make_transformed_dataset(predictions)

    def _fit_predict(
        self, predictions: SparkDataset, pipes: Sequence[SparkMLPipeline]
    ) -> Tuple[SparkDataset, Sequence[SparkMLPipeline]]:
        raise NotImplementedError()

    # noinspection PyMethodMayBeStatic
    def split_models(self, predictions: SparkDataset, pipes: Sequence[SparkMLPipeline]) -> List[Tuple[str, int, int]]:
        """Split predictions by single model prediction datasets.

        Args:
            predictions: Dataset with predictions.
            pipes: ml pipelines to be associated with the predictions

        Returns:
            Each tuple in the list is:
            - prediction column name
            - corresponding model index (in the pipe)
            - corresponding pipe index

        """
        return [
            (ml_algo.prediction_feature, j, i)
            for i, pipe in enumerate(pipes)
            for j, ml_algo in enumerate(pipe.ml_algos)
        ]

    def _set_metadata(self, predictions: SparkDataset, pipes: Sequence[SparkMLPipeline]):
        assert len(predictions.roles) > 0, "The predictions dataset should have at least one column"
        self._input_roles = copy(predictions.roles)
        self._pred_role = list(predictions.roles.values())[0]
        self._output_roles = {self._single_prediction_col_name: self._pred_role}

        if isinstance(self._pred_role, NumericVectorOrArrayRole):
            self._outp_dim = self._pred_role.size
        else:
            self._outp_dim = 1
        self._outp_prob = predictions.task.name in ["binary", "multiclass"]
        self._score = predictions.task.get_dataset_metric()
        self._task = predictions.task

    def score(self, dataset: SparkDataset) -> float:
        """Score metric for blender.

        Args:
            dataset: Blended predictions dataset.

        Returns:
            Metric value.

        """
        return self._score(dataset, True)


class SparkBestModelSelector(SparkBlender, WeightedBlender):
    """Select best single model from level.

    Drops pipes that are not used in calc best model.
    Works in general case (even on some custom things)
    and most efficient on inference.
    Perform worse than other on tables,
    specially if some of models was terminated by timer.

    """

    def _fit_predict(
        self, predictions: SparkDataset, pipes: Sequence[SparkMLPipeline]
    ) -> Tuple[SparkDataset, Sequence[SparkMLPipeline]]:
        """Simple fit - just take one best.

        Args:
            predictions: Sequence of datasets with predictions.
            pipes: Sequence of pipelines.

        Returns:
            Single prediction dataset and Sequence of pruned pipelines.

        """
        splitted_models_and_pipes = self.split_models(predictions, pipes)

        best_pred = None
        best_pred_col = None
        best_pipe_idx = 0
        best_model_idx = 0
        best_score = -np.inf

        for pred_col, mod, pipe in splitted_models_and_pipes:
            # pred_ds = self._make_single_pred_ds(predictions, pred_col)
            pred_ds = cast(SparkDataset, predictions[:, [pred_col]])
            score = self.score(pred_ds)

            if score > best_score:
                best_pipe_idx = pipe
                best_model_idx = mod
                best_score = score
                best_pred = pred_ds
                best_pred_col = pred_col

        best_pipe = pipes[best_pipe_idx]
        best_pipe.ml_algos = [best_pipe.ml_algos[best_model_idx]]

        self._transformer = Pipeline(
            stages=[
                SQLTransformer(
                    statement=f"SELECT *, {best_pred_col} AS {self._single_prediction_col_name} FROM __THIS__"
                ),
                ColumnsSelectorTransformer(
                    name=f"{type(self)}",
                    input_cols=[self._single_prediction_col_name],
                    optional_cols=predictions.service_columns,
                ),
            ]
        ).fit(best_pred.data)

        self._output_roles = {self._single_prediction_col_name: best_pred.roles[best_pred_col]}

        out_ds = best_pred.empty()
        out_ds.set_data(
            self._transformer.transform(best_pred.data),
            list(self._output_roles.keys()),
            self._output_roles,
            name=type(self).__name__,
        )

        return out_ds, [best_pipe]


class SparkWeightedBlender(SparkBlender, WeightedBlender):
    """Weighted Blender based on coord descent, optimize task metric directly.

    Weight sum eq. 1.
    Good blender for tabular data,
    even if some predictions are NaN (ex. timeout).
    Model with low weights will be pruned.

    """

    def __init__(
        self,
        max_iters: int = 5,
        max_inner_iters: int = 7,
        max_nonzero_coef: float = 0.05,
    ):
        SparkBlender.__init__(self)
        WeightedBlender.__init__(self, max_iters, max_inner_iters, max_nonzero_coef)
        self._predictions_dataset: Optional[SparkDataset] = None

    def _get_weighted_pred(
        self,
        splitted_preds: Sequence[str],
        wts: Optional[np.ndarray],
        remove_splitted_preds_cols: Optional[List[str]] = None,
    ) -> SparkDataset:
        avr = self._build_avr_transformer(splitted_preds, wts, remove_splitted_preds_cols)
        vsh = self._build_vector_size_hint(self._single_prediction_col_name, self._pred_role)

        weighted_preds_sdf = PipelineModel(stages=[avr, vsh]).transform(self._predictions_dataset.data)

        wpreds_sds = self._predictions_dataset.empty()
        wpreds_sds.set_data(
            weighted_preds_sdf, list(self.output_roles.keys()), self.output_roles, name=type(self).__name__
        )

        return wpreds_sds

    def _build_avr_transformer(
        self,
        splitted_preds: Sequence[str],
        wts: Optional[np.ndarray],
        remove_splitted_preds_cols: Optional[List[str]] = None,
    ) -> AveragingTransformer:
        remove_cols = list(splitted_preds)

        if remove_splitted_preds_cols is not None:
            remove_cols.extend(remove_splitted_preds_cols)

        return AveragingTransformer(
            task_name=self._task.name,
            input_cols=list(splitted_preds),
            output_col=self._single_prediction_col_name,
            remove_cols=remove_cols,
            convert_to_array_first=True,
            weights=(wts * len(wts)).tolist(),
            dim_num=self._outp_dim,
        )

    def _fit_predict(
        self, predictions: SparkDataset, pipes: Sequence[SparkMLPipeline]
    ) -> Tuple[SparkDataset, Sequence[SparkMLPipeline]]:
        self._predictions_dataset = predictions

        sm = self.split_models(predictions, pipes)
        pred_cols = [pred_col for pred_col, _, _ in sm]
        pipe_idx = np.array([pidx for _, _, pidx in sm])

        wts = self._optimize(pred_cols)

        reweighted_pred_cols = [x for (x, w) in zip(pred_cols, wts) if w > 0]
        removed_cols = [x for x in pred_cols if x not in reweighted_pred_cols]
        _, self.wts = self._prune_pipe(pipes, wts, pipe_idx)
        pipes = cast(Sequence[SparkMLPipeline], pipes)

        self._transformer = self._build_avr_transformer(
            reweighted_pred_cols, self.wts, remove_splitted_preds_cols=removed_cols
        )
        outp = self._get_weighted_pred(reweighted_pred_cols, self.wts, remove_splitted_preds_cols=removed_cols)

        return outp, pipes


class SparkMeanBlender(SparkBlender):
    """Simple average level predictions.

    Works only with TabularDatasets.
    Doesn't require target to fit.
    No pruning.

    """

    def _fit_predict(
        self, predictions: SparkDataset, pipes: Sequence[SparkMLPipeline]
    ) -> Tuple[SparkDataset, Sequence[SparkMLPipeline]]:
        pred_cols = [pred_col for pred_col, _, _ in self.split_models(predictions, pipes)]

        avr = AveragingTransformer(
            task_name=predictions.task.name,
            input_cols=pred_cols,
            output_col=self._single_prediction_col_name,
            remove_cols=pred_cols,
            convert_to_array_first=not (predictions.task.name == "reg"),
            dim_num=self._outp_dim,
        )
        vsh = self._build_vector_size_hint(self._single_prediction_col_name, self._pred_role)

        self._transformer = PipelineModel(stages=[avr, vsh])

        df = self._transformer.transform(predictions.data)

        if predictions.task.name in ["binary", "multiclass"]:
            assert isinstance(self._pred_role, NumericVectorOrArrayRole)
            output_role = NumericVectorOrArrayRole(
                self._pred_role.size,
                "MeanBlend_{{}}",
                dtype=np.float32,
                prob=self._outp_prob,
                is_vector=self._pred_role.is_vector,
            )
        else:
            output_role = NumericRole(np.float32, prob=self._outp_prob)

        roles = {self._single_prediction_col_name: output_role}
        pred_ds = predictions.empty()
        pred_ds.set_data(df, list(roles.keys()), roles, name=type(self).__name__)

        self._output_roles = copy(roles)

        return pred_ds, pipes



"""Base class for selection pipelines."""
from abc import ABC
from copy import copy
from typing import List
from typing import Optional
from typing import cast

from lightautoml.dataset.base import LAMLDataset
from lightautoml.dataset.base import RolesDict
from lightautoml.pipelines.selection.base import ComposedSelector
from lightautoml.pipelines.selection.base import EmptySelector
from lightautoml.pipelines.selection.base import ImportanceEstimator
from lightautoml.pipelines.selection.base import SelectionPipeline
from lightautoml.validation.base import TrainValidIterator
from pandas import Series
from pyspark.ml import Transformer

from sparklightautoml.dataset.base import SparkDataset
from sparklightautoml.pipelines.base import TransformerInputOutputRoles
from sparklightautoml.pipelines.features.base import SparkFeaturesPipeline
from sparklightautoml.utils import NoOpTransformer
from sparklightautoml.validation.base import SparkBaseTrainValidIterator
from sparklightautoml.validation.base import SparkSelectionPipeline


class SparkImportanceEstimator(ImportanceEstimator, ABC):
    def __init__(self):
        super(SparkImportanceEstimator, self).__init__()


class SparkSelectionPipelineWrapper(SparkSelectionPipeline, TransformerInputOutputRoles):
    def __init__(self, sel_pipe: SelectionPipeline):
        # assert not sel_pipe.is_fitted, "Cannot work with prefitted SelectionPipeline"
        self._validate_sel_pipe(sel_pipe)

        self._sel_pipe = sel_pipe
        self._service_columns = None
        self._is_fitted = False
        self._input_roles: Optional[RolesDict] = None
        self._output_roles: Optional[RolesDict] = None
        self._feature_pipeline = cast(SparkFeaturesPipeline, self._sel_pipe.features_pipeline)
        self._service_columns: Optional[List[str]] = None
        super().__init__()

    def _validate_sel_pipe(self, sel_pipe: SelectionPipeline):
        selectors = sel_pipe.selectors if isinstance(sel_pipe, ComposedSelector) else [sel_pipe]
        for selp in selectors:
            msg = (
                f"SelectionPipeline should either be EmptySelector or have SparkFeaturePipeline as "
                f"features_pipeline, but it is {type(selp)} and have {type(selp.features_pipeline)}"
            )
            assert (
                isinstance(selp, EmptySelector)
                or isinstance(selp, BugFixSelectionPipelineWrapper)
                or isinstance(selp.features_pipeline, SparkFeaturesPipeline)
            ), msg

    def _build_transformer(self, *args, **kwargs) -> Optional[Transformer]:
        if not self._sel_pipe.is_fitted:
            return None

        # we cannot select columns on infer time
        # because ml pipes are executed sequentially during predict
        # applying selector on predict may lead to loss of columns
        # required for the subsequent ml pipes
        return NoOpTransformer(name=f"{type(self)}")

    @property
    def input_roles(self) -> Optional[RolesDict]:
        return self._input_roles

    @property
    def output_roles(self) -> Optional[RolesDict]:
        return self._output_roles

    @property
    def is_fitted(self) -> bool:
        return self._is_fitted

    @property
    def selected_features(self) -> List[str]:
        return self._sel_pipe.selected_features

    @property
    def in_features(self) -> List[str]:
        return self._sel_pipe.in_features

    @property
    def dropped_features(self) -> List[str]:
        return self._sel_pipe.dropped_features

    def perform_selection(self, train_valid: Optional[TrainValidIterator]):
        self._sel_pipe.perform_selection(train_valid)

    def fit(self, train_valid: SparkBaseTrainValidIterator):
        self._service_columns = train_valid.train.service_columns

        if not self._sel_pipe.is_fitted:
            self._sel_pipe.fit(train_valid)
        self._is_fitted = True

        self._input_roles = copy(train_valid.train.roles)
        self._output_roles = {
            feat: role for feat, role in self._input_roles.items() if feat in self._sel_pipe.selected_features
        }
        self._service_columns = train_valid.train.service_columns

    def select(self, dataset: SparkDataset) -> SparkDataset:
        return cast(SparkDataset, self._sel_pipe.select(dataset))

    def map_raw_feature_importances(self, raw_importances: Series):
        return self._sel_pipe.map_raw_feature_importances(raw_importances)

    def get_features_score(self):
        return self._sel_pipe.get_features_score()

    def _get_service_columns(self) -> List[str]:
        return self._service_columns


class BugFixSelectionPipelineWrapper(SelectionPipeline):
    def __init__(self, instance: SelectionPipeline):
        super(BugFixSelectionPipelineWrapper, self).__init__()
        self._instance = instance

    def fit(self, train_valid: TrainValidIterator):
        self._instance.fit(train_valid)

    def map_raw_feature_importances(self, raw_importances: Series):
        return self._instance.map_raw_feature_importances(raw_importances)

    def get_features_score(self):
        return self._instance.get_features_score()

    @property
    def is_fitted(self) -> bool:
        return self._instance.is_fitted

    @property
    def selected_features(self) -> List[str]:
        return self._instance.selected_features

    @property
    def in_features(self) -> List[str]:
        return self._instance.in_features

    @property
    def dropped_features(self) -> List[str]:
        return self._instance.dropped_features

    def perform_selection(self, train_valid: Optional[TrainValidIterator]):
        self._instance.perform_selection(train_valid)

    def select(self, dataset: LAMLDataset) -> LAMLDataset:
        selected_features = copy(self._instance.selected_features)
        # Add features that forces input
        sl_set = set(selected_features)
        roles = dataset.roles
        for col in (x for x in dataset.features if x not in sl_set):
            if roles[col].force_input:
                if col not in sl_set:
                    selected_features.append(col)

        return dataset[:, selected_features]



import logging

from copy import deepcopy
from typing import Callable
from typing import Optional
from typing import Tuple
from typing import Union

import optuna

from lightautoml.ml_algo.tuning.optuna import OptunaTuner
from lightautoml.ml_algo.tuning.optuna import TunableAlgo
from lightautoml.validation.base import HoldoutIterator

from sparklightautoml.computations.base import ComputationsSession
from sparklightautoml.computations.base import ComputationsSettings
from sparklightautoml.computations.builder import build_computations_manager
from sparklightautoml.computations.sequential import SequentialComputationsManager
from sparklightautoml.computations.utils import deecopy_tviter_without_dataset
from sparklightautoml.dataset.base import SparkDataset
from sparklightautoml.ml_algo.base import SparkTabularMLAlgo
from sparklightautoml.validation.base import SparkBaseTrainValidIterator


logger = logging.getLogger(__name__)


class ParallelOptunaTuner(OptunaTuner):
    def __init__(
        self,
        timeout: Optional[int] = 1000,
        n_trials: Optional[int] = 100,
        direction: Optional[str] = "maximize",
        fit_on_holdout: bool = True,
        random_state: int = 42,
        parallelism: int = 1,
        computations_manager: Optional[ComputationsSettings] = None,
    ):
        super().__init__(timeout, n_trials, direction, fit_on_holdout, random_state)
        self._parallelism = parallelism
        self._computations_manager = build_computations_manager(computations_settings=computations_manager)
        self._session: Optional[ComputationsSession] = None

    def fit(
        self, ml_algo: SparkTabularMLAlgo, train_valid_iterator: Optional[SparkBaseTrainValidIterator] = None
    ) -> Tuple[Optional[SparkTabularMLAlgo], Optional[SparkDataset]]:
        """Tune model.

        Args:
            ml_algo: Algo that is tuned.
            train_valid_iterator: Classic cv-iterator.

        Returns:
            Tuple (None, None) if an optuna exception raised
            or ``fit_on_holdout=True`` and ``train_valid_iterator`` is
            not :class:`~lightautoml.validation.base.HoldoutIterator`.
            Tuple (MlALgo, preds_ds) otherwise.

        """
        assert not ml_algo.is_fitted, "Fitted algo cannot be tuned."

        estimated_tuning_time = ml_algo.timer.estimate_tuner_time(len(train_valid_iterator))
        if estimated_tuning_time:
            estimated_tuning_time = max(estimated_tuning_time, 1)
            self._upd_timeout(estimated_tuning_time)

        logger.info(
            f"Start hyperparameters optimization for \x1b[1m{ml_algo._name}\x1b[0m ... Time budget is {self.timeout:.2f} secs"
        )

        metric_name = train_valid_iterator.train.task.get_dataset_metric().name
        ml_algo = deepcopy(ml_algo)

        flg_new_iterator = False
        if self._fit_on_holdout and type(train_valid_iterator) != HoldoutIterator:
            train_valid_iterator = train_valid_iterator.convert_to_holdout_iterator()
            flg_new_iterator = True

        def update_trial_time(study: optuna.study.Study, trial: optuna.trial.FrozenTrial):
            """Callback for number of iteration with time cut-off.

            Args:
                study: Optuna study object.
                trial: Optuna trial object.

            """
            ml_algo.mean_trial_time = study.trials_dataframe()["duration"].mean().total_seconds()
            self.estimated_n_trials = min(self.n_trials, self.timeout // ml_algo.mean_trial_time)

            logger.info3(
                f"\x1b[1mTrial {len(study.trials)}\x1b[0m with hyperparameters {trial.params} scored {trial.value} in {trial.duration}"
            )

        try:
            self._optimize(ml_algo, train_valid_iterator, update_trial_time)

            # need to update best params here
            self._best_params = self.study.best_params
            ml_algo.params = self._best_params

            logger.info(f"Hyperparameters optimization for \x1b[1m{ml_algo._name}\x1b[0m completed")
            logger.info2(
                f"The set of hyperparameters \x1b[1m{self._best_params}\x1b[0m\n achieve {self.study.best_value:.4f} {metric_name}"
            )

            if flg_new_iterator:
                # if tuner was fitted on holdout set we dont need to save train results
                return None, None

            preds_ds = ml_algo.fit_predict(train_valid_iterator)

            return ml_algo, preds_ds
        except optuna.exceptions.OptunaError:
            return None, None
        except:
            logger.error("Error during parameters optimization", exc_info=True)
            raise

    def _optimize(
        self,
        ml_algo: SparkTabularMLAlgo,
        train_valid_iterator: SparkBaseTrainValidIterator,
        update_trial_time: Callable[[optuna.study.Study, optuna.trial.FrozenTrial], None],
    ):
        sampler = optuna.samplers.TPESampler(seed=self.random_state)
        self.study = optuna.create_study(direction=self.direction, sampler=sampler)

        # prepare correct ml_algo to run with optuna
        cm = ml_algo.computations_manager
        trial_ml_algo = deepcopy(ml_algo)
        ml_algo.computations_manager = cm
        trial_ml_algo.persist_output_dataset = False

        with self._computations_manager.session(train_valid_iterator.train) as self._session:
            # _get objective will use self._session
            self.study.optimize(
                func=self._get_objective(
                    ml_algo=trial_ml_algo,
                    estimated_n_trials=self.estimated_n_trials,
                    train_valid_iterator=train_valid_iterator,
                ),
                n_jobs=self._parallelism,
                n_trials=self.n_trials,
                timeout=self.timeout,
                callbacks=[update_trial_time],
            )

        self._session = None

    def _get_objective(
        self, ml_algo: TunableAlgo, estimated_n_trials: int, train_valid_iterator: SparkBaseTrainValidIterator
    ) -> Callable[[optuna.trial.Trial], Union[float, int]]:
        assert isinstance(ml_algo, SparkTabularMLAlgo)

        def objective(trial: optuna.trial.Trial) -> float:
            with self._session.allocate() as slot:
                assert slot.dataset is not None
                _ml_algo = deepcopy(ml_algo)
                _ml_algo.computations_manager = SequentialComputationsManager(
                    num_tasks=slot.num_tasks, num_threads_per_executor=slot.num_threads_per_executor
                )
                tv_iter = deecopy_tviter_without_dataset(train_valid_iterator)
                tv_iter.train = slot.dataset

                optimization_search_space = _ml_algo.optimization_search_space

                if not optimization_search_space:
                    optimization_search_space = _ml_algo._get_default_search_spaces(
                        suggested_params=_ml_algo.init_params_on_input(tv_iter),
                        estimated_n_trials=estimated_n_trials,
                    )

                if callable(optimization_search_space):
                    _ml_algo.params = optimization_search_space(
                        trial=trial,
                        optimization_search_space=optimization_search_space,
                        suggested_params=_ml_algo.init_params_on_input(tv_iter),
                    )
                else:
                    _ml_algo.params = self._sample(
                        trial=trial,
                        optimization_search_space=optimization_search_space,
                        suggested_params=_ml_algo.init_params_on_input(tv_iter),
                    )

                logger.debug(f"Sampled ml_algo params: {_ml_algo.params}")

                output_dataset = _ml_algo.fit_predict(train_valid_iterator=tv_iter)

                return _ml_algo.score(output_dataset)

        return objective



import logging
import os
import uuid
import warnings

from abc import abstractmethod
from dataclasses import dataclass
from typing import Dict
from typing import List
from typing import Optional
from typing import Union
from typing import cast

from sparklightautoml.dataset.base import PersistableDataFrame
from sparklightautoml.dataset.base import PersistenceLevel
from sparklightautoml.dataset.base import PersistenceManager
from sparklightautoml.dataset.base import SparkDataset
from sparklightautoml.utils import JobGroup
from sparklightautoml.utils import get_current_session


logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class _PersistablePair:
    level: PersistenceLevel
    pdf: PersistableDataFrame


class BasePersistenceManager(PersistenceManager):
    """
    Abstract implementation of base persistence functionality, including registering and de-registering
    what have been requested to persist/un-persist
    """

    def __init__(self, parent: Optional["PersistenceManager"] = None):
        self._uid = str(uuid.uuid4())
        self._persistence_registry: Dict[str, _PersistablePair] = dict()
        self._parent = parent
        self._children: Dict[str, "PersistenceManager"] = dict()

    @property
    def uid(self) -> str:
        return self._uid

    @property
    def children(self) -> List["PersistenceManager"]:
        return list(self._children.values())

    @property
    def datasets(self) -> List[SparkDataset]:
        return [pair.pdf.to_dataset() for pair in self._persistence_registry.values()]

    @property
    def all_datasets(self) -> List[SparkDataset]:
        return [*self.datasets, *(ds for child in self.children for ds in child.all_datasets)]

    def persist(
        self, dataset: Union[SparkDataset, PersistableDataFrame], level: PersistenceLevel = PersistenceLevel.REGULAR
    ) -> PersistableDataFrame:
        persisted_dataframe = (
            self.to_persistable_dataframe(dataset)
            if isinstance(dataset, SparkDataset)
            else cast(PersistableDataFrame, dataset)
        )

        logger.debug(
            f"Manager {self._uid}: " f"persisting dataset (uid={dataset.uid}, name={dataset.name}) with level {level}."
        )

        if persisted_dataframe.uid in self._persistence_registry:
            persisted_pair = self._persistence_registry[persisted_dataframe.uid]

            logger.debug(
                f"Manager {self._uid}: " f"the dataset (uid={dataset.uid}, name={dataset.name}) is already persisted."
            )

            if persisted_pair.level != level:
                warnings.warn(
                    f"Asking to persist an already persisted dataset "
                    f"(uid={persisted_pair.pdf.uid}, name={persisted_pair.pdf.name}) "
                    f"but with different level. Will do nothing. "
                    f"Current level {persisted_pair.level}, asked level {level}.",
                    RuntimeWarning,
                )

            return persisted_pair.pdf

        self._persistence_registry[persisted_dataframe.uid] = _PersistablePair(
            level, self._persist(persisted_dataframe, level)
        )

        logger.debug(f"Manager {self._uid}: the dataset (uid={dataset.uid}, name={dataset.name}) has been persisted.")

        return self._persistence_registry[persisted_dataframe.uid].pdf

    def unpersist(self, uid: str):
        persisted_pair = self._persistence_registry.get(uid, None)

        if not persisted_pair:
            logger.debug(f"Manager {self._uid} unpersist: the dataset (uid={uid}) is not persisted yet. Nothing to do.")
            return

        logger.debug(f"Manager {self._uid}: unpersisting dataset (uid={uid}, name={persisted_pair.pdf.name}).")

        self._unpersist(persisted_pair.pdf)

        del self._persistence_registry[persisted_pair.pdf.uid]

        logger.debug(
            f"Manager {self._uid}: " f"the dataset (uid={uid}, name={persisted_pair.pdf.name}) has been unpersisted."
        )

    def unpersist_children(self):
        logger.info(f"Manager {self._uid}: unpersisting children.")

        for child in self._children:
            child.unpersist_all()
        self._children = dict()

        logger.debug(f"Manager {self._uid}: children have been unpersisted.")

    def unpersist_all(self):
        logger.info(f"Manager {self._uid}: unpersisting everything.")

        self.unpersist_children()

        uids = list(self._persistence_registry.keys())

        for uid in uids:
            self.unpersist(uid)

        logger.debug(f"Manager {self._uid}: everything has been unpersisted.")

    def child(self) -> "PersistenceManager":
        # TODO: SLAMA - Bugfix
        logger.info(f"Manager {self._uid}: producing a child.")
        a_child = self._create_child()
        self._children[a_child.uid] = a_child
        logger.info(f"Manager {self._uid}: the child (uid={a_child.uid}) has been produced.")
        return a_child

    def remove_child(self, child: Union["PersistenceManager", str]):
        uid = child.uid if isinstance(child, PersistenceManager) else child

        if uid not in self._children:
            logger.warning(f"Not found child with uid {uid} to delete from parent {self._uid}")
            return

        del self._children[uid]

    def is_persisted(self, pdf: PersistableDataFrame) -> bool:
        return pdf.uid in self._persistence_registry

    @abstractmethod
    def _persist(self, pdf: PersistableDataFrame, level: PersistenceLevel) -> PersistableDataFrame:
        ...

    @abstractmethod
    def _unpersist(self, pdf: PersistableDataFrame):
        ...

    @abstractmethod
    def _create_child(self) -> PersistenceManager:
        ...


class PlainCachePersistenceManager(BasePersistenceManager):
    """
    Manager that uses Spark .cache() / .persist() methods
    """

    def __init__(self, parent: Optional["PersistenceManager"] = None, prune_history: bool = False):
        super().__init__(parent)
        self._prune_history = prune_history

    def _persist(self, pdf: PersistableDataFrame, level: PersistenceLevel) -> PersistableDataFrame:
        logger.debug(
            f"Manager {self._uid}: " f"caching and materializing the dataset (uid={pdf.uid}, name={pdf.name})."
        )

        with JobGroup(
            "Persisting", f"{type(self)} caching df (uid={pdf.uid}, name={pdf.name})", pdf.sdf.sql_ctx.sparkSession
        ):
            df = (
                pdf.sdf.sql_ctx.sparkSession.createDataFrame(pdf.sdf.rdd, schema=pdf.sdf.schema)
                if self._prune_history
                else pdf.sdf
            )
            ds = df.cache()
            ds.write.mode("overwrite").format("noop").save()

        logger.debug(f"Manager {self._uid}: " f"caching succeeded for the dataset (uid={pdf.uid}, name={pdf.name}).")

        return PersistableDataFrame(ds, pdf.uid, pdf.callback, pdf.base_dataset)

    def _unpersist(self, pdf: PersistableDataFrame):
        pdf.sdf.unpersist()

    def _create_child(self) -> PersistenceManager:
        return PlainCachePersistenceManager(self)


class LocalCheckpointPersistenceManager(BasePersistenceManager):
    """
    Manager that uses Spark .localCheckpoint() method
    """

    def __init__(self, parent: Optional["PersistenceManager"] = None):
        super().__init__(parent)

    def _persist(self, pdf: PersistableDataFrame, level: PersistenceLevel) -> PersistableDataFrame:
        logger.debug(
            f"Manager {self._uid}: " f"making a local checkpoint for the dataset (uid={pdf.uid}, name={pdf.name})."
        )

        with JobGroup(
            "Persisting",
            f"{type(self)} local checkpointing of df (uid={pdf.uid}, name={pdf.name})",
            pdf.sdf.sql_ctx.sparkSession,
        ):
            ds = pdf.sdf.localCheckpoint()

        logger.debug(
            f"Manager {self._uid}: "
            f"the local checkpoint has been made for the dataset (uid={pdf.uid}, name={pdf.name})."
        )
        return PersistableDataFrame(ds, pdf.uid, pdf.callback, pdf.base_dataset)

    def _unpersist(self, pdf: PersistableDataFrame):
        pdf.sdf.unpersist()

    def _create_child(self) -> PersistenceManager:
        return LocalCheckpointPersistenceManager(self)


class BucketedPersistenceManager(BasePersistenceManager):
    """
    Manager that uses Spark Warehouse folder to store bucketed datasets (.bucketBy ... .sortBy ... .saveAsTable)
    To make such storing reliable, one should set 'spark.sql.warehouse.dir' to HDFS or other reliable storage.
    """

    def __init__(
        self,
        bucketed_datasets_folder: str,
        bucket_nums: int = 100,
        parent: Optional["PersistenceManager"] = None,
        no_unpersisting: bool = False,
    ):
        super().__init__(parent)
        self._bucket_nums = bucket_nums
        self._bucketed_datasets_folder = bucketed_datasets_folder
        self._no_unpersisting = no_unpersisting

    def _persist(self, pdf: PersistableDataFrame, level: PersistenceLevel) -> PersistableDataFrame:
        spark = get_current_session()
        name = self._build_name(pdf)
        # TODO: SLAMA join - need to identify correct setting  for bucket_nums if it is not provided
        path = self._build_path(name)
        logger.debug(
            f"Manager {self._uid}: making a bucketed table "
            f"for the dataset (uid={pdf.uid}, name={pdf.name}) with name {name} on path {path}."
        )

        with JobGroup(
            "Persisting",
            f"{type(self)} saving bucketed table of df (uid={pdf.uid}, name={pdf.name}). Table path: {path}",
            pdf.sdf.sql_ctx.sparkSession,
        ):
            # If we directly put path in .saveAsTable(...), than Spark will create an external table
            # that cannot be physically deleted with .sql("DROP TABLE <name>")
            # Without stating the path, Spark will create a managed table
            (
                pdf.sdf.repartition(self._bucket_nums, SparkDataset.ID_COLUMN)
                .write.mode("overwrite")
                .bucketBy(self._bucket_nums, SparkDataset.ID_COLUMN)
                .sortBy(SparkDataset.ID_COLUMN)
                .saveAsTable(name, format="parquet")
            )
        ds = spark.table(name)

        logger.debug(
            f"Manager {self._uid}: the bucketed table has been made "
            f"for the dataset (uid={pdf.uid}, name={pdf.name}) with name {name} on path {path}."
        )
        return PersistableDataFrame(ds, pdf.uid, pdf.callback, pdf.base_dataset)

    def _unpersist(self, pdf: PersistableDataFrame):
        if self._no_unpersisting:
            return

        name = self._build_name(pdf)
        path = self._build_path(name)
        logger.debug(
            f"Manager {self._uid}: removing the bucketed table "
            f"for the dataset (uid={pdf.uid}, name={pdf.name}) with name {name} on path {path}."
        )

        get_current_session().sql(f"DROP TABLE {name}")

        logger.debug(
            f"Manager {self._uid}: the bucketed table has been removed"
            f"for the dataset (uid={pdf.uid}, name={pdf.name}) with name {name} on path {path}."
        )

    def _create_child(self) -> PersistenceManager:
        return BucketedPersistenceManager(self._bucketed_datasets_folder, self._bucket_nums, self)

    def _build_path(self, name: str) -> str:
        return os.path.join(self._bucketed_datasets_folder, f"{name}.parquet")

    @staticmethod
    def _build_name(pdf: PersistableDataFrame):
        return f"{pdf.name}_{pdf.uid}".replace("-", "__")


class CompositePersistenceManager(BasePersistenceManager):
    """
    Universal composite manager that can combine other manager to apply different
    storing strategies on different levels.

    For BucketedPersistenceManager all unpersisting operations are delayed until the end of automl processing,
    due to possible loss of source for downstream persistence manager if they don't use external storage and files.
    """

    def __init__(
        self, level2manager: Dict[PersistenceLevel, PersistenceManager], parent: Optional["PersistenceManager"] = None
    ):
        super().__init__(parent)
        self._level2manager = level2manager
        self._force = False

    def _persist(self, pdf: PersistableDataFrame, level: PersistenceLevel) -> PersistableDataFrame:
        assert level in self._level2manager, (
            f"Cannot process level {level} because the corresponding manager has not been set. "
            f"Only the following levels are supported: {list(self._level2manager.keys())}."
        )

        persisted_on_levels = [
            lvl for lvl, manager in self._level2manager.items() if manager.is_persisted(pdf) and lvl != level
        ]

        assert len(persisted_on_levels) == 0, (
            f"Unable to persist with the required level {level}, because the dataset has been already persisted "
            f"with different levels: {persisted_on_levels}"
        )

        return self._level2manager[level].persist(pdf, level)

    def _unpersist(self, pdf: PersistableDataFrame):
        for lvl, manager in self._level2manager.items():
            if manager.is_persisted(pdf):
                # we cannot unpersist anything in BucketedPersistenceManager
                # when it is working with other managers of different types
                # If an other persistence manager is of PlainCachePersistenceManager type
                # and is used to persist a dataset referencing some dataset persisted with BucketedPersistenceManager,
                # than after unpersisting with deleting for that referenced dataset will lead to failing of calculation
                # for the first dataset due to inability to obtain info about the source file
                # All such persisted datasets should be unpersisted with unpersist_all method in the end of a scenario
                if not isinstance(manager, BucketedPersistenceManager) or self._force:
                    manager.unpersist(pdf.uid)
                break

    def _create_child(self) -> PersistenceManager:
        level2managers = {lvl: manager.child() for lvl, manager in self._level2manager.items()}
        return CompositePersistenceManager(level2managers, self)

    def unpersist_all(self):
        self._force = True
        super(CompositePersistenceManager, self).unpersist_all()
        self._force = False


class CompositePlainCachePersistenceManager(CompositePersistenceManager):
    """
    Combines PlainCache on READER and REGULAR levels with bucketing on CHECKPOINT level.
    """

    def __init__(self, bucketed_datasets_folder: str, bucket_nums: int):
        super(CompositePlainCachePersistenceManager, self).__init__(
            {
                PersistenceLevel.READER: BucketedPersistenceManager(
                    bucketed_datasets_folder=bucketed_datasets_folder, bucket_nums=bucket_nums, no_unpersisting=True
                ),
                PersistenceLevel.REGULAR: PlainCachePersistenceManager(),
                PersistenceLevel.CHECKPOINT: PlainCachePersistenceManager(),
            }
        )


class CompositeBucketedPersistenceManager(CompositePersistenceManager):
    """
    Combines bucketing on READER and CHECKPOINT levels with PlainCache on REGULAR level.
    """

    def __init__(self, bucketed_datasets_folder: str, bucket_nums: int):
        super(CompositeBucketedPersistenceManager, self).__init__(
            {
                PersistenceLevel.READER: BucketedPersistenceManager(
                    bucketed_datasets_folder=bucketed_datasets_folder, bucket_nums=bucket_nums, no_unpersisting=True
                ),
                PersistenceLevel.REGULAR: PlainCachePersistenceManager(prune_history=False),
                PersistenceLevel.CHECKPOINT: BucketedPersistenceManager(
                    bucketed_datasets_folder=bucketed_datasets_folder, bucket_nums=bucket_nums
                ),
            }
        )



from contextlib import contextmanager
from typing import Callable
from typing import List
from typing import Optional

from sparklightautoml.computations.base import ComputationSlot
from sparklightautoml.computations.base import ComputationsManager
from sparklightautoml.computations.base import ComputationsSession
from sparklightautoml.computations.base import R
from sparklightautoml.computations.base import T
from sparklightautoml.dataset.base import SparkDataset


class SequentialComputationsSession(ComputationsSession):
    def __init__(
        self,
        dataset: Optional[SparkDataset] = None,
        num_tasks: Optional[int] = None,
        num_threads_per_executor: Optional[int] = None,
    ):
        super(SequentialComputationsSession, self).__init__()
        self._dataset = dataset
        self._num_tasks = num_tasks
        self._num_threads_per_executor = num_threads_per_executor

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    @contextmanager
    def allocate(self) -> ComputationSlot:
        yield ComputationSlot(
            "0", self._dataset, num_tasks=self._num_tasks, num_threads_per_executor=self._num_threads_per_executor
        )

    def map_and_compute(self, func: Callable[[R], T], tasks: List[R]) -> List[T]:
        return [func(task) for task in tasks]

    def compute(self, tasks: List[Callable[[], T]]) -> List[T]:
        return [task() for task in tasks]


class SequentialComputationsManager(ComputationsManager):
    def __init__(self, num_tasks: Optional[int] = None, num_threads_per_executor: Optional[int] = None):
        super(SequentialComputationsManager, self).__init__()
        self._dataset: Optional[SparkDataset] = None
        self._num_tasks = num_tasks
        self._num_threads_per_executor = num_threads_per_executor

    @property
    def parallelism(self) -> int:
        return 1

    def session(self, dataset: Optional[SparkDataset] = None) -> SequentialComputationsSession:
        return SequentialComputationsSession(
            dataset, num_tasks=self._num_tasks, num_threads_per_executor=self._num_threads_per_executor
        )



import logging
import os
import socket
import time
import warnings

from contextlib import contextmanager
from datetime import datetime
from logging import Logger
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import cast

import pyspark

from py4j.java_gateway import java_import
from pyspark import RDD
from pyspark.ml import Estimator
from pyspark.ml import Transformer
from pyspark.ml.common import inherit_doc
from pyspark.ml.param import Param
from pyspark.ml.param import Params
from pyspark.ml.param import TypeConverters
from pyspark.ml.param.shared import HasInputCols
from pyspark.ml.param.shared import HasOutputCols
from pyspark.ml.pipeline import PipelineModel
from pyspark.ml.pipeline import PipelineSharedReadWrite
from pyspark.ml.util import DefaultParamsReadable
from pyspark.ml.util import DefaultParamsReader
from pyspark.ml.util import DefaultParamsWritable
from pyspark.ml.util import DefaultParamsWriter
from pyspark.ml.util import MLReader
from pyspark.ml.util import MLWriter
from pyspark.sql import SparkSession


VERBOSE_LOGGING_FORMAT = "%(asctime)s %(threadName)s %(levelname)s %(module)s %(filename)s:%(lineno)d %(message)s"

logger = logging.getLogger(__name__)


@contextmanager
def spark_session(
    session_args: Optional[dict] = None, master: str = "local[]", wait_secs_after_the_end: Optional[int] = None
) -> SparkSession:
    """
    Args:
        session_args: additional arguments to be add to SparkSession using .config() method
        master: address of the master
            to run locally - "local[1]"

            to run on spark cluster - "spark://node4.bdcl:7077"
            (Optionally set the driver host to a correct hostname .config("spark.driver.host", "node4.bdcl"))

        wait_secs_after_the_end: amount of seconds to wait before stoping SparkSession and thus web UI.

    Returns:
        SparkSession to be used and that is stopped upon exiting this context manager
    """

    if not session_args:
        spark_sess_builder = (
            SparkSession.builder.appName("SPARK-LAMA-app")
            .master(master)
            .config("spark.jars", "jars/spark-lightautoml_2.12-0.1.jar")
            .config("spark.jars.packages", "com.microsoft.azure:synapseml_2.12:0.9.5")
            .config("spark.jars.repositories", "https://mmlspark.azureedge.net/maven")
            .config("spark.sql.shuffle.partitions", "16")
            .config("spark.kryoserializer.buffer.max", "512m")
            .config("spark.driver.cores", "4")
            .config("spark.driver.memory", "16g")
            .config("spark.cores.max", "16")
            .config("spark.executor.instances", "4")
            .config("spark.executor.memory", "16g")
            .config("spark.executor.cores", "4")
            .config("spark.memory.fraction", "0.6")
            .config("spark.memory.storageFraction", "0.5")
            .config("spark.sql.autoBroadcastJoinThreshold", "100MB")
            .config("spark.sql.execution.arrow.pyspark.enabled", "true")
        )
    else:
        spark_sess_builder = SparkSession.builder.appName("SPARK-LAMA-app")
        for arg, value in session_args.items():
            spark_sess_builder = spark_sess_builder.config(arg, value)

        if "spark.master" in session_args and not session_args["spark.master"].startswith("local["):
            local_ip_address = socket.gethostbyname(socket.gethostname())
            logger.info(f"Using IP address for spark driver: {local_ip_address}")
            spark_sess_builder = spark_sess_builder.config("spark.driver.host", local_ip_address)

    spark_sess = spark_sess_builder.getOrCreate()

    logger.info(f"Spark WebUI url: {spark_sess.sparkContext.uiWebUrl}")

    try:
        yield spark_sess
    finally:
        logger.info(
            f"The session is ended. Sleeping {wait_secs_after_the_end if wait_secs_after_the_end else 0} "
            f"secs until stop the spark session."
        )
        if wait_secs_after_the_end:
            time.sleep(wait_secs_after_the_end)
        spark_sess.stop()


@contextmanager
def log_exec_time(name: Optional[str] = None, write_log=True):
    # Add file handler for INFO
    if write_log:
        file_handler_info = logging.FileHandler(f"/tmp/{name}_log.log.log", mode="a")
        file_handler_info.setFormatter(logging.Formatter("%(message)s"))
        file_handler_info.setLevel(logging.INFO)
        logger.addHandler(file_handler_info)

    start = datetime.now()

    yield

    end = datetime.now()
    duration = (end - start).total_seconds()

    msg = f"Exec time of {name}: {duration}" if name else f"Exec time: {duration}"
    logger.warning(msg)


# log_exec_time() class to return elapsed time value
class log_exec_timer:
    def __init__(self, name: Optional[str] = None):
        self.name = name
        self._start = None
        self._duration = None

    def __enter__(self):
        self._start = datetime.now()
        return self

    def __exit__(self, typ, value, traceback):
        self._duration = (datetime.now() - self._start).total_seconds()
        msg = f"Exec time of {self.name}: {self._duration}" if self.name else f"Exec time: {self._duration}"
        logger.info(msg)

    @property
    def duration(self):
        return self._duration


SparkDataFrame = pyspark.sql.DataFrame


def get_current_session() -> SparkSession:
    return SparkSession.builder.getOrCreate()


def get_cached_df_through_rdd(df: SparkDataFrame, name: Optional[str] = None) -> Tuple[SparkDataFrame, RDD]:
    rdd = df.rdd
    cached_rdd = rdd.setName(name).cache() if name else rdd.cache()
    cached_df = df.sql_ctx.createDataFrame(cached_rdd, df.schema)
    return cached_df, cached_rdd


def logging_config(level: int = logging.INFO, log_filename: str = "/var/log/lama.log") -> dict:
    return {
        "version": 1,
        "disable_existing_loggers": True,
        "formatters": {
            "verbose": {"format": VERBOSE_LOGGING_FORMAT},
            "simple": {"format": "%(asctime)s %(levelname)s %(message)s"},
        },
        "handlers": {
            "console": {"level": "DEBUG", "class": "logging.StreamHandler", "formatter": "verbose"},
            "file": {
                "level": "DEBUG",
                "class": "logging.FileHandler",
                "formatter": "verbose",
                "filename": log_filename,
            },
        },
        "loggers": {
            "lightautoml": {
                "handlers": ["console", "file"],
                "propagate": True,
                "level": level,
            },
            "sparklightautoml": {
                "handlers": ["console", "file"],
                "level": level,
                "propagate": False,
            },
            "lightautoml.ml_algo": {
                "handlers": ["console", "file"],
                "level": level,
                "propagate": False,
            },
        },
    }


def cache(df: SparkDataFrame) -> SparkDataFrame:
    if not df.is_cached:
        df = df.cache()
    return df


def warn_if_not_cached(df: SparkDataFrame):
    if not df.is_cached:
        warnings.warn(
            "Attempting to calculate shape on not cached dataframe. " "It may take too much time.", RuntimeWarning
        )


class ColumnsSelectorTransformer(
    Transformer, HasInputCols, HasOutputCols, DefaultParamsWritable, DefaultParamsReadable
):
    """
    Makes selection input columns from input dataframe.
    """

    optionalCols = Param(
        Params._dummy(), "optionalCols", "optional column names.", typeConverter=TypeConverters.toListString
    )

    transformOnlyFirstTime = Param(
        Params._dummy(),
        "transformOnlyFirstTime",
        "whatever to transform only once or each time",
        typeConverter=TypeConverters.toBoolean,
    )

    _alreadyTransformed = Param(
        Params._dummy(),
        "_alreadyTransformed",
        "is it first time to transform or not",
        typeConverter=TypeConverters.toBoolean,
    )

    def __init__(
        self,
        name: Optional[str] = None,
        input_cols: Optional[List[str]] = None,
        optional_cols: Optional[List[str]] = None,
        transform_only_first_time: bool = False,
    ):
        super().__init__()
        input_cols = input_cols if input_cols else []
        optional_cols = optional_cols if optional_cols else []
        assert (
            len(set(input_cols).intersection(set(optional_cols))) == 0
        ), "Input columns and optional columns cannot intersect"

        self._name = name
        self.set(self.inputCols, input_cols)
        self.set(self.optionalCols, optional_cols)
        self.set(self.outputCols, input_cols)
        self.set(self.transformOnlyFirstTime, transform_only_first_time)
        self.set(self._alreadyTransformed, False)

    def get_optional_cols(self) -> List[str]:
        return self.getOrDefault(self.optionalCols)

    def _transform(self, dataset: SparkDataFrame) -> SparkDataFrame:
        logger.debug(f"In {type(self)}. Name: {self._name}. Columns: {sorted(dataset.columns)}")

        if not self.getOrDefault(self.transformOnlyFirstTime) or not self.getOrDefault(self._alreadyTransformed):
            ds_cols = set(dataset.columns)
            present_opt_cols = [c for c in self.get_optional_cols() if c in ds_cols]
            input_cols = self._treat_columns_pattern(dataset, self.getInputCols())
            opt_input_cols = self._treat_columns_pattern(dataset, present_opt_cols)
            dataset = dataset.select([*input_cols, *opt_input_cols])
            self.set(self._alreadyTransformed, True)
            self.set(self.outputCols, input_cols)

        logger.debug(f"Out {type(self)}. Name: {self._name}. Columns: {sorted(dataset.columns)}")
        return dataset

    @staticmethod
    def _treat_columns_pattern(df: SparkDataFrame, cols: List[str]) -> List[str]:
        def treat(col: str):
            if col.endswith("*"):
                pattern = col[:-1]
                return [c for c in df.columns if c.startswith(pattern)]
            return [col]

        cols = [cc for c in cols for cc in treat(c)]
        return cols


class NoOpTransformer(Transformer, DefaultParamsWritable, DefaultParamsReadable):
    def __init__(self, name: Optional[str] = None):
        super().__init__()
        self._name = name

    def _transform(self, dataset: SparkDataFrame) -> SparkDataFrame:
        logger.debug(f"In {type(self)}. Name: {self._name}. Columns: {sorted(dataset.columns)}")
        return dataset


@inherit_doc
class WrappingSelectingPipelineModelWriter(MLWriter):
    """
    (Private) Specialization of :py:class:`MLWriter` for :py:class:`PipelineModel` types
    """

    def __init__(self, instance: "WrappingSelectingPipelineModel"):
        super(WrappingSelectingPipelineModelWriter, self).__init__()
        self.instance = instance

    def saveImpl(self, path: str):
        stages = self.instance.stages
        PipelineSharedReadWrite.validateStages(stages)

        stageUids = [stage.uid for stage in stages]
        jsonParams = {
            "stageUids": stageUids,
            "language": "Python",
            "instance_params": {param.name: value for param, value in self.instance.extractParamMap().items()},
        }
        DefaultParamsWriter.saveMetadata(self.instance, path, self.sc, paramMap=jsonParams)
        stagesDir = os.path.join(path, "stages")
        for index, stage in enumerate(stages):
            stage.write().save(PipelineSharedReadWrite.getStagePath(stage.uid, index, len(stages), stagesDir))


@inherit_doc
class WrappingSelectionPipelineModelReader(MLReader):
    """
    (Private) Specialization of :py:class:`MLReader` for :py:class:`PipelineModel` types
    """

    def __init__(self, cls):
        super(WrappingSelectionPipelineModelReader, self).__init__()
        self.cls = cls

    def load(self, path):
        metadata = DefaultParamsReader.loadMetadata(path, self.sc)

        uid, stages = PipelineSharedReadWrite.load(metadata, self.sc, path)
        instance_params_map = cast(Dict, metadata["paramMap"]["instance_params"])

        wspm = WrappingSelectingPipelineModel(stages=stages)
        wspm._resetUid(uid)
        for param_name, value in instance_params_map.items():
            param = wspm.getParam(param_name)
            wspm.set(param, value)

        return wspm


class WrappingSelectingPipelineModel(PipelineModel, HasInputCols):
    name = Param(Params._dummy(), "name", "name.", typeConverter=TypeConverters.toString)

    optionalCols = Param(
        Params._dummy(), "optionalCols", "optional column names.", typeConverter=TypeConverters.toListString
    )

    def __init__(
        self,
        stages: List[Transformer],
        input_columns: Optional[List[str]] = None,
        optional_columns: Optional[List[str]] = None,
        name: Optional[str] = None,
    ):
        super().__init__(stages)
        self.set(self.inputCols, input_columns or [])
        self.set(self.optionalCols, optional_columns or [])
        self.set(self.name, name or "")

    def _transform(self, dataset: SparkDataFrame) -> SparkDataFrame:
        cstr = ColumnsSelectorTransformer(
            name=f"{type(self).__name__}({self.getOrDefault(self.name)})",
            input_cols=list({*dataset.columns, *self.getInputCols()}),
            optional_cols=self.getOrDefault(self.optionalCols),
        )
        ds = super()._transform(dataset)
        return cstr.transform(ds)

    def write(self):
        """Returns an MLWriter instance for this ML instance."""
        return WrappingSelectingPipelineModelWriter(self)

    @classmethod
    def read(cls):
        """Returns an MLReader instance for this class."""
        return WrappingSelectionPipelineModelReader(cls)


class Cacher(Estimator):
    _cacher_dict: Dict[str, SparkDataFrame] = dict()

    @classmethod
    def get_dataset_by_key(cls, key: str) -> Optional[SparkDataFrame]:
        return cls._cacher_dict.get(key, None)

    @classmethod
    def release_cache_by_key(cls, key: str):
        df = cls._cacher_dict.pop(key, None)
        if df is not None:
            df.unpersist()
            del df

    @property
    def dataset(self) -> SparkDataFrame:
        """Returns chached dataframe"""
        return self._cacher_dict[self._key]

    def __init__(self, key: str):
        super().__init__()
        self._key = key
        self._dataset: Optional[SparkDataFrame] = None

    def _fit(self, dataset):
        logger.info(f"Cacher {self._key} (RDD Id: {dataset.rdd.id()}). Starting to materialize data.")

        # using local checkpoints
        # ds = dataset.localCheckpoint(eager=True)

        # # using plain caching
        # ds = get_current_session().createDataFrame(dataset.rdd, schema=dataset.schema).cache()
        ds = dataset.cache()
        ds.write.mode("overwrite").format("noop").save()

        logger.info(
            f"Cacher {self._key} (RDD Id: {ds.rdd.id()}, Column nums: {len(ds.columns)}). "
            f"Finished data materialization."
        )

        previous_ds = self._cacher_dict.get(self._key, None)
        if previous_ds is not None and ds != previous_ds:
            logger.info(f"Removing cache for key: {self._key} (RDD Id: {previous_ds.rdd.id()}).")
            previous_ds.unpersist()
            del previous_ds

        self._cacher_dict[self._key] = ds

        return NoOpTransformer(name=f"cacher_{self._key}")


class EmptyCacher(Cacher):
    def __init__(self, key: str):
        super().__init__(key)
        self._dataset: Optional[SparkDataFrame] = None

    @property
    def dataset(self) -> SparkDataFrame:
        return self._dataset

    def _fit(self, dataset):
        self._dataset = dataset
        return NoOpTransformer(name=f"empty_cacher_{self._key}")


def log_exception(logger: Logger):
    def wrap(func):
        def wrapped_f(*args, **kwargs):
            try:
                result = func(*args, **kwargs)
            except Exception as ex:
                logger.error("Error wrapper caught error", exc_info=True)
                raise ex
            return result

        return wrapped_f

    return wrap


@contextmanager
def JobGroup(group_id: str, description: str, spark: SparkSession):
    sc = spark.sparkContext
    sc.setJobGroup(group_id, description)
    yield
    sc._jsc.clearJobGroup()


# noinspection PyProtectedMember,PyUnresolvedReferences
def create_directory(path: str, spark: SparkSession, exists_ok: bool = False):
    java_import(spark._jvm, "org.apache.hadoop.fs.Path")
    java_import(spark._jvm, "java.net.URI")
    java_import(spark._jvm, "org.apache.hadoop.fs.FileSystem")

    juri = spark._jvm.Path(path).toUri()
    jpath = spark._jvm.Path(juri.getPath())
    jscheme = spark._jvm.URI(f"{juri.getScheme()}://{juri.getAuthority() or ''}/") if juri.getScheme() else None
    fs = (
        spark._jvm.FileSystem.get(jscheme, spark._jsc.hadoopConfiguration())
        if jscheme
        else spark._jvm.FileSystem.get(spark._jsc.hadoopConfiguration())
    )

    if not fs.exists(jpath):
        fs.mkdirs(jpath)
    elif not exists_ok:
        raise FileExistsError(f"The path already exists: {path}")



import logging
import os

from copy import copy
from copy import deepcopy
from typing import Iterable
from typing import List
from typing import Optional
from typing import Sequence
from typing import Tuple

import numpy as np

from lightautoml.automl.presets.base import upd_params
from lightautoml.automl.presets.utils import plot_pdp_with_distribution
from lightautoml.pipelines.selection.base import ComposedSelector
from lightautoml.pipelines.selection.importance_based import ImportanceCutoffSelector
from lightautoml.pipelines.selection.importance_based import (
    ModelBasedImportanceEstimator,
)
from lightautoml.pipelines.selection.permutation_importance_based import (
    NpIterativeFeatureSelector,
)
from lightautoml.reader.tabular_batch_generator import ReadableToDf
from lightautoml.utils.timer import TaskTimer
from pyspark.ml import Transformer
from pyspark.sql import SparkSession
from pyspark.sql import Window
from pyspark.sql import functions as sf
from pyspark.sql.types import DateType
from pyspark.sql.types import StringType
from tqdm import tqdm

from sparklightautoml.automl.base import ReadableIntoSparkDf
from sparklightautoml.automl.blend import SparkWeightedBlender
from sparklightautoml.automl.presets.base import SparkAutoMLPreset
from sparklightautoml.automl.presets.utils import calc_feats_permutation_imps
from sparklightautoml.automl.presets.utils import replace_dayofweek_in_date
from sparklightautoml.automl.presets.utils import replace_month_in_date
from sparklightautoml.automl.presets.utils import replace_year_in_date
from sparklightautoml.computations.builder import AutoMLComputationsSettings
from sparklightautoml.computations.builder import ComputationsManagerFactory
from sparklightautoml.dataset.base import PersistenceManager
from sparklightautoml.dataset.base import SparkDataset
from sparklightautoml.dataset.persistence import PlainCachePersistenceManager
from sparklightautoml.ml_algo.boost_lgbm import SparkBoostLGBM
from sparklightautoml.ml_algo.linear_pyspark import SparkLinearLBFGS
from sparklightautoml.ml_algo.tuning.parallel_optuna import ParallelOptunaTuner
from sparklightautoml.pipelines.features.lgb_pipeline import SparkLGBAdvancedPipeline
from sparklightautoml.pipelines.features.lgb_pipeline import SparkLGBSimpleFeatures
from sparklightautoml.pipelines.features.linear_pipeline import SparkLinearFeatures
from sparklightautoml.pipelines.ml.nested_ml_pipe import SparkNestedTabularMLPipeline
from sparklightautoml.pipelines.selection.base import BugFixSelectionPipelineWrapper
from sparklightautoml.pipelines.selection.base import SparkSelectionPipelineWrapper
from sparklightautoml.pipelines.selection.permutation_importance_based import (
    SparkNpPermutationImportanceEstimator,
)
from sparklightautoml.reader.base import SparkToSparkReader
from sparklightautoml.tasks.base import SparkTask
from sparklightautoml.utils import SparkDataFrame


logger = logging.getLogger(__name__)
base_dir = os.path.dirname(__file__)


class SparkTabularAutoML(SparkAutoMLPreset):
    """
    Spark version of :class:`TabularAutoML <lightautoml.automl.presets.tabular_presets.TabularAutoML>`.
    Represent high level entity of spark lightautoml. Use this class to create automl instance.

    Example:

        >>> automl = SparkTabularAutoML(
        >>>     spark=spark,
        >>>     task=SparkTask('binary'),
        >>>     general_params={"use_algos": [["lgb"]]},
        >>>     lgb_params={'use_single_dataset_mode': True},
        >>>     reader_params={"cv": cv, "advanced_roles": False}
        >>> )
        >>> oof_predictions = automl.fit_predict(
        >>>     train_data,
        >>>     roles=roles
        >>> )
    """

    _default_config_path = "tabular_config.yml"

    # set initial runtime rate guess for first level models
    _time_scores = {
        "lgb": 1,
        "lgb_tuned": 3,
        "linear_l2": 0.7,
        "cb": 2,
        "cb_tuned": 6,
    }

    def __init__(
        self,
        spark: SparkSession,
        task: SparkTask,
        timeout: int = 3600,
        memory_limit: int = 16,
        cpu_limit: int = 4,
        gpu_ids: Optional[str] = "all",
        timing_params: Optional[dict] = None,
        config_path: Optional[str] = None,
        general_params: Optional[dict] = None,
        reader_params: Optional[dict] = None,
        read_csv_params: Optional[dict] = None,
        nested_cv_params: Optional[dict] = None,
        tuning_params: Optional[dict] = None,
        selection_params: Optional[dict] = None,
        lgb_params: Optional[dict] = None,
        cb_params: Optional[dict] = None,
        linear_l2_params: Optional[dict] = None,
        gbm_pipeline_params: Optional[dict] = None,
        linear_pipeline_params: Optional[dict] = None,
        persistence_manager: Optional[PersistenceManager] = None,
        computation_settings: AutoMLComputationsSettings = ("no_parallelism", -1),
    ):
        if config_path is None:
            config_path = os.path.join(base_dir, self._default_config_path)

        self._computation_managers_factory = (
            computation_settings
            if isinstance(computation_settings, ComputationsManagerFactory)
            else ComputationsManagerFactory(computation_settings)
        )

        super().__init__(
            task,
            timeout,
            memory_limit,
            cpu_limit,
            gpu_ids,
            timing_params,
            config_path,
            self._computation_managers_factory.get_ml_pipelines_manager(),
        )

        self._persistence_manager = persistence_manager or PlainCachePersistenceManager()

        self._spark = spark
        # upd manual params
        for name, param in zip(
            [
                "general_params",
                "reader_params",
                "read_csv_params",
                "nested_cv_params",
                "tuning_params",
                "selection_params",
                "lgb_params",
                "cb_params",
                "linear_l2_params",
                "gbm_pipeline_params",
                "linear_pipeline_params",
            ],
            [
                general_params,
                reader_params,
                read_csv_params,
                nested_cv_params,
                tuning_params,
                selection_params,
                lgb_params,
                cb_params,
                linear_l2_params,
                gbm_pipeline_params,
                linear_pipeline_params,
            ],
        ):
            if param is None:
                param = {}
            self.__dict__[name] = upd_params(self.__dict__[name], param)

    def infer_auto_params(self, train_data: SparkDataFrame, multilevel_avail: bool = False):
        # infer optuna tuning iteration based on dataframe len
        if self.tuning_params["max_tuning_iter"] == "auto":
            if not train_data.is_cached:
                self.tuning_params["max_tuning_iter"] = 5

            length = train_data.count()

            if length < 10000:
                self.tuning_params["max_tuning_iter"] = 100
            elif length < 30000:
                self.tuning_params["max_tuning_iter"] = 50
            elif length < 100000:
                self.tuning_params["max_tuning_iter"] = 10
            else:
                self.tuning_params["max_tuning_iter"] = 5

        if self.general_params["use_algos"] == "auto":
            self.general_params["use_algos"] = [["lgb", "lgb_tuned", "linear_l2", "cb", "cb_tuned"]]
            if self.task.name == "multiclass" and multilevel_avail:
                self.general_params["use_algos"].append(["linear_l2", "lgb"])

        if not self.general_params["nested_cv"]:
            self.nested_cv_params["cv"] = 1

    def get_time_score(self, n_level: int, model_type: str, nested: Optional[bool] = None):
        if nested is None:
            nested = self.general_params["nested_cv"]

        score = self._time_scores[model_type]

        mult = 1
        if nested:
            if self.nested_cv_params["n_folds"] is not None:
                mult = self.nested_cv_params["n_folds"]
            else:
                mult = self.nested_cv_params["cv"]

        if n_level > 1:
            mult *= 0.8 if self.general_params["skip_conn"] else 0.1

        score = score * mult

        # lower score for catboost on gpu
        if model_type in ["cb", "cb_tuned"] and self.cb_params["default_params"]["task_type"] == "GPU":
            score *= 0.5
        return score

    def get_selector(self, n_level: Optional[int] = 1) -> SparkSelectionPipelineWrapper:
        selection_params = self.selection_params
        # lgb_params
        lgb_params = deepcopy(self.lgb_params)
        lgb_params["default_params"] = {
            **lgb_params["default_params"],
            **{"featureFraction": 1},
        }

        mode = selection_params["mode"]

        # create pre selection based on mode
        pre_selector = None
        if mode > 0:
            # if we need selector - define model
            # timer will be useful to estimate time for next gbm runs
            time_score = self.get_time_score(n_level, "lgb", False)

            sel_timer_0 = self.timer.get_task_timer("lgb", time_score)
            selection_feats = SparkLGBSimpleFeatures()

            selection_gbm = SparkBoostLGBM(timer=sel_timer_0, **lgb_params)
            selection_gbm.set_prefix("Selector")

            if selection_params["importance_type"] == "permutation":
                importance = SparkNpPermutationImportanceEstimator()
            else:
                importance = ModelBasedImportanceEstimator()

            pre_selector = ImportanceCutoffSelector(
                selection_feats,
                selection_gbm,
                importance,
                cutoff=selection_params["cutoff"],
                fit_on_holdout=selection_params["fit_on_holdout"],
            )
            if mode == 2:
                time_score = self.get_time_score(n_level, "lgb", False)

                sel_timer_1 = self.timer.get_task_timer("lgb", time_score)
                selection_feats = SparkLGBSimpleFeatures()
                selection_gbm = SparkBoostLGBM(timer=sel_timer_1, **lgb_params)
                selection_gbm.set_prefix("Selector")

                importance = SparkNpPermutationImportanceEstimator(
                    computations_settings=self._computation_managers_factory.get_selector_manager()
                )

                extra_selector = NpIterativeFeatureSelector(
                    selection_feats,
                    selection_gbm,
                    importance,
                    feature_group_size=selection_params["feature_group_size"],
                    max_features_cnt_in_result=selection_params["max_features_cnt_in_result"],
                )

                pre_selector = ComposedSelector([pre_selector, extra_selector])

        pre_selector = BugFixSelectionPipelineWrapper(pre_selector)
        return SparkSelectionPipelineWrapper(pre_selector)

    def get_linear(
        self, n_level: int = 1, pre_selector: Optional[SparkSelectionPipelineWrapper] = None
    ) -> SparkNestedTabularMLPipeline:
        # linear model with l2
        time_score = self.get_time_score(n_level, "linear_l2")
        linear_l2_timer = self.timer.get_task_timer("reg_l2", time_score)
        linear_l2_params = {**self.linear_l2_params}

        linear_l2_model = SparkLinearLBFGS(
            timer=linear_l2_timer,
            computations_settings=self._computation_managers_factory.get_linear_manager(),
            **linear_l2_params,
        )
        linear_l2_feats = SparkLinearFeatures(output_categories=True, **self.linear_pipeline_params)

        linear_l2_pipe = SparkNestedTabularMLPipeline(
            [linear_l2_model],
            force_calc=True,
            pre_selection=pre_selector,
            features_pipeline=linear_l2_feats,
            computations_settings=self._computation_managers_factory.get_ml_algo_manager(),
            **self.nested_cv_params,
        )
        return linear_l2_pipe

    def get_gbms(
        self,
        keys: Sequence[str],
        n_level: int = 1,
        pre_selector: Optional[SparkSelectionPipelineWrapper] = None,
    ):
        gbm_feats = SparkLGBAdvancedPipeline(**self.gbm_pipeline_params)

        ml_algos = []
        force_calc = []
        for key, force in zip(keys, [True, False, False, False]):
            tuned = "_tuned" in key
            algo_key = key.split("_")[0]
            time_score = self.get_time_score(n_level, key)
            gbm_timer = self.timer.get_task_timer(algo_key, time_score)
            gbm_model, lgb_params = self._get_boosting_model(algo_key, gbm_timer)

            if tuned:
                gbm_model.set_prefix("Tuned")

                tuner_comp_manager = self._computation_managers_factory.get_tuning_manager()

                gbm_tuner = ParallelOptunaTuner(
                    n_trials=self.tuning_params["max_tuning_iter"],
                    timeout=self.tuning_params["max_tuning_time"],
                    fit_on_holdout=self.tuning_params["fit_on_holdout"],
                    parallelism=tuner_comp_manager.parallelism,
                    computations_manager=tuner_comp_manager,
                )
                gbm_model = (gbm_model, gbm_tuner)

            ml_algos.append(gbm_model)
            force_calc.append(force)

        gbm_pipe = SparkNestedTabularMLPipeline(
            ml_algos,
            force_calc,
            pre_selection=pre_selector,
            features_pipeline=gbm_feats,
            computations_settings=self._computation_managers_factory.get_ml_algo_manager(),
            **self.nested_cv_params,
        )

        return gbm_pipe

    def _get_boosting_model(self, algo_key: str, gbm_timer: Optional[TaskTimer] = None):
        if algo_key == "lgb":
            lgb_params = {**self.lgb_params}

            gbm_model = SparkBoostLGBM(
                timer=gbm_timer,
                computations_settings=self._computation_managers_factory.get_lgb_manager(),
                **lgb_params,
            )
            return gbm_model, lgb_params
        elif algo_key == "cb":
            raise NotImplementedError("Not supported yet")
        else:
            raise ValueError("Wrong algo key")

    def create_automl(self, **fit_args):
        """Create basic automl instance.

        Args:
            **fit_args: Contain all information needed for creating automl.

        """
        train_data = fit_args["train_data"]
        multilevel_avail = fit_args["valid_data"] is None and fit_args["cv_iter"] is None

        self.infer_auto_params(train_data, multilevel_avail)
        reader = SparkToSparkReader(task=self.task, **self.reader_params)

        pre_selector = self.get_selector()

        levels = []

        for n, names in enumerate(self.general_params["use_algos"]):
            lvl = []
            # regs
            if "linear_l2" in names:
                selector = None
                if "linear_l2" in self.selection_params["select_algos"] and (
                    self.general_params["skip_conn"] or n == 0
                ):
                    selector = pre_selector
                lvl.append(self.get_linear(n + 1, selector))

            gbm_models = [
                x for x in ["lgb", "lgb_tuned", "cb", "cb_tuned"] if x in names and x.split("_")[0] in self.task.losses
            ]

            if len(gbm_models) > 0:
                selector = None
                if "gbm" in self.selection_params["select_algos"] and (self.general_params["skip_conn"] or n == 0):
                    selector = pre_selector
                lvl.append(self.get_gbms(gbm_models, n + 1, selector))

            levels.append(lvl)

        # blend everything
        blender = SparkWeightedBlender(max_nonzero_coef=self.general_params["weighted_blender_max_nonzero_coef"])

        # initialize
        self._initialize(
            reader,
            levels,
            skip_conn=self.general_params["skip_conn"],
            blender=blender,
            return_all_predictions=self.general_params["return_all_predictions"],
            timer=self.timer,
        )

    def _get_read_csv_params(self):
        try:
            cols_to_read = self.reader.used_features
            numeric_dtypes = {
                x: self.reader.roles[x].dtype for x in self.reader.roles if self.reader.roles[x].name == "Numeric"
            }
        except AttributeError:
            cols_to_read = []
            numeric_dtypes = {}
        # cols_to_read is empty if reader is not fitted
        if len(cols_to_read) == 0:
            cols_to_read = None

        read_csv_params = copy(self.read_csv_params)
        read_csv_params = {
            **read_csv_params,
            **{"usecols": cols_to_read, "dtype": numeric_dtypes},
        }

        return read_csv_params

    def fit_predict(
        self,
        train_data: ReadableIntoSparkDf,
        roles: Optional[dict] = None,
        train_features: Optional[Sequence[str]] = None,
        cv_iter: Optional[Iterable] = None,
        valid_data: Optional[ReadableIntoSparkDf] = None,
        valid_features: Optional[Sequence[str]] = None,
        log_file: str = None,
        verbose: int = 0,
        persistence_manager: Optional[PersistenceManager] = None,
    ) -> SparkDataset:
        """Fit and get prediction on validation dataset.

        Almost same as :meth:`lightautoml.automl.base.AutoML.fit_predict`.

        Additional features - working with different data formats.
        Supported now:

            - Path to ``.csv``, ``.parquet``, ``.feather`` files.
            - :class:`~numpy.ndarray`, or dict of :class:`~numpy.ndarray`.
              For example, ``{'data': X...}``. In this case,
              roles are optional, but `train_features`
              and `valid_features` required.
            - :class:`pandas.DataFrame`.

        Args:
            train_data: Dataset to train.
            roles: Roles dict.
            train_features: Optional features names, if can't
              be inferred from `train_data`.
            cv_iter: Custom cv-iterator. For example,
              :class:`~lightautoml.validation.np_iterators.TimeSeriesIterator`.
            valid_data: Optional validation dataset.
            valid_features: Optional validation dataset features
              if cannot be inferred from `valid_data`.
            verbose: Controls the verbosity: the higher, the more messages.
                <1  : messages are not displayed;
                >=1 : the computation process for layers is displayed;
                >=2 : the information about folds processing is also displayed;
                >=3 : the hyperparameters optimization process is also displayed;
                >=4 : the training process for every algorithm is displayed;
            log_file: Filename for writing logging messages. If log_file is specified,
            the messages will be saved in a the file. If the file exists, it will be overwritten.

        Returns:
            Dataset with predictions. Call ``.data`` to get predictions array.

        """
        # roles may be none in case of train data is set {'data': np.ndarray, 'target': np.ndarray ...}
        self.set_logfile(log_file)

        if roles is None:
            roles = {}
        train = self._read_data(train_data, train_features)
        # if upd_roles:
        #     roles = {**roles, **upd_roles}
        if valid_data is not None:
            valid_data = self._read_data(valid_data, valid_features)

        oof_pred = super().fit_predict(
            train,
            roles=roles,
            cv_iter=cv_iter,
            valid_data=valid_data,
            verbose=verbose,
            persistence_manager=persistence_manager,
        )

        return oof_pred

    def get_feature_scores(
        self,
        calc_method: str = "fast",
        data: Optional[ReadableIntoSparkDf] = None,
        features_names: Optional[Sequence[str]] = None,
        silent: bool = True,
    ):
        if calc_method == "fast":
            for level in self.levels:
                for pipe in level:
                    fi = pipe.pre_selection.get_features_score()
                    if fi is not None:
                        used_feats = set(self.collect_used_feats())
                        fi = fi.reset_index()
                        fi.columns = ["Feature", "Importance"]
                        return fi[fi["Feature"].map(lambda x: x in used_feats)]

            else:
                if not silent:
                    logger.info2("No feature importances to show. Please use another calculation method")
                return None

        if calc_method != "accurate":
            if not silent:
                logger.info2(
                    "Unknown calc_method. "
                    + "Currently supported methods for feature importances calculation are 'fast' and 'accurate'."
                )
            return None

        if data is None:
            if not silent:
                logger.info2("Data parameter is not setup for accurate calculation method. Aborting...")
            return None

        read_csv_params = self._get_read_csv_params()
        data = self._read_data(data, features_names, read_csv_params)
        used_feats = self.collect_used_feats()
        fi = calc_feats_permutation_imps(
            self,
            used_feats,
            data,
            self.task.get_dataset_metric(),
            silent=silent,
        )
        return fi

    @staticmethod
    def get_histogram(data: SparkDataFrame, column: str, n_bins: int) -> Tuple[List, np.ndarray]:
        assert n_bins >= 2, "n_bins must be equal 2 or more"
        bin_edges, counts = (
            data.select(sf.col(column).cast("double"))
            .where(sf.col(column).isNotNull())
            .rdd.map(lambda x: x[0])
            .histogram(n_bins)
        )
        bin_edges = np.array(bin_edges)
        return counts, bin_edges

    @staticmethod
    def get_pdp_data_numeric_feature(
        df: SparkDataFrame,
        feature_name: str,
        model: Transformer,
        prediction_col: str,
        n_bins: int,
        ice_fraction: float = 1.0,
        ice_fraction_seed: int = 42,
    ) -> Tuple[List, List, List]:
        """Returns `grid`, `ys` and `counts` calculated on input numeric column to plot PDP.

        Args:
            df (SparkDataFrame): Spark DataFrame with `feature_name` column
            feature_name (str): feature column name
            model (PipelineModel): Spark Pipeline Model
            prediction_col (str): prediction column to be created by the `model`
            n_bins (int): The number of bins to produce. Raises exception if n_bins < 2.
            ice_fraction (float, optional): What fraction of the input dataframe will be used to make predictions.
                Useful for very large dataframe. Defaults to 1.0.
            ice_fraction_seed (int, optional): Seed for `ice_fraction`. Defaults to 42.

        Returns:
            Tuple[List, List, List]:
            `grid` is list of categories,
            `ys` is list of predictions by category,
            `counts` is numbers of values by category
        """
        counts, bin_edges = SparkTabularAutoML.get_histogram(df, feature_name, n_bins)
        grid = (bin_edges[:-1] + bin_edges[1:]) / 2
        ys = []
        sample_df = (
            df.select(*[c for c in df.columns if c != feature_name])
            .sample(fraction=ice_fraction, seed=ice_fraction_seed)
            .cache()
        )
        for i in tqdm(grid):
            # replace feature column values with constant
            sdf = sample_df.select("*", sf.lit(i).alias(feature_name))

            # infer via transformer
            preds = model.transform(sdf)
            # TODO: SPARK-LAMA remove this line after passing the "prediction_col" parameter
            prediction_col = next(c for c in preds.columns if c.startswith("prediction"))
            preds = np.array(preds.select(prediction_col).collect())
            # when preds.shape is (n, 1, k) we change it to (n, k),
            # where n is number of rows and k is number of classes
            if len(preds.shape) == 3:
                preds = np.squeeze(preds, axis=1)
            ys.append(preds)

        sample_df.unpersist()

        return grid, ys, counts

    @staticmethod
    def get_pdp_data_categorical_feature(
        df: SparkDataFrame,
        feature_name: str,
        model: Transformer,
        prediction_col: str,
        n_top_cats: int,
        ice_fraction: float = 1.0,
        ice_fraction_seed: int = 42,
    ) -> Tuple[List, List, List]:
        """Returns `grid`, `ys` and `counts` calculated on input categorical column to plot PDP.

        Args:
            df (SparkDataFrame): Spark DataFrame with `feature_name` column
            feature_name (str): feature column name
            model (PipelineModel): Spark Pipeline Model
            prediction_col (str): prediction column to be created by the `model`
            n_top_cats (int): param to selection top n categories
            ice_fraction (float, optional): What fraction of the input dataframe will be used to make predictions.
                Useful for very large dataframe. Defaults to 1.0.
            ice_fraction_seed (int, optional): Seed for `ice_fraction`. Defaults to 42.

        Returns:
            Tuple[List, List, List]:
            `grid` is list of categories,
            `ys` is list of predictions by category,
            `counts` is numbers of values by category
        """
        feature_cnt = (
            df.where(sf.col(feature_name).isNotNull()).groupBy(feature_name).count().orderBy(sf.desc("count")).collect()
        )
        grid = [row[feature_name] for row in feature_cnt[:n_top_cats]]
        counts = [row["count"] for row in feature_cnt[:n_top_cats]]
        ys = []
        sample_df = (
            df.select(*[c for c in df.columns if c != feature_name])
            .sample(fraction=ice_fraction, seed=ice_fraction_seed)
            .cache()
        )
        for i in tqdm(grid):
            sdf = sample_df.select("*", sf.lit(i).alias(feature_name))
            preds = model.transform(sdf)
            # TODO: SPARK-LAMA remove this line after passing the "prediction_col" parameter
            prediction_col = next(c for c in preds.columns if c.startswith("prediction"))
            preds = np.array(preds.select(prediction_col).collect())
            # when preds.shape is (n, 1, k) we change it to (n, k),
            # where n is number of rows and k is number of classes
            if len(preds.shape) == 3:
                preds = np.squeeze(preds, axis=1)
            ys.append(preds)
        if len(feature_cnt) > n_top_cats:
            # unique other categories
            unique_other_categories = [row[feature_name] for row in feature_cnt[n_top_cats:]]

            # get non-top categories, natural distributions is important here
            w = Window().orderBy(sf.lit("A"))  # window without sorting
            other_categories_collection = (
                df.select(feature_name)
                .filter(sf.col(feature_name).isin(unique_other_categories))
                .select(sf.row_number().over(w).alias("row_num"), feature_name)
                .collect()
            )

            # dict with key=%row number% and value=%category%
            other_categories_dict = {x["row_num"]: x[feature_name] for x in other_categories_collection}
            max_row_num = len(other_categories_collection)

            def get_category_by_row_num(row_num):
                remainder = row_num % max_row_num
                if remainder == 0:
                    key = max_row_num
                else:
                    key = remainder
                return other_categories_dict[key]

            get_category_udf = sf.udf(get_category_by_row_num, StringType())

            # add row number to main dataframe and exclude feature_name column
            sdf = sample_df.select("*", sf.row_number().over(w).alias("row_num"))

            all_columns_except_row_num = [f for f in sdf.columns if f != "row_num"]
            feature_col = get_category_udf(sf.col("row_num")).alias(feature_name)
            # exclude row number from dataframe
            # and add back feature_name column filled with other categories same distribution
            sdf = sdf.select(*all_columns_except_row_num, feature_col)

            preds = model.transform(sdf)
            preds = np.array(preds.select(prediction_col).collect())
            # when preds.shape is (n, 1, k) we change it to (n, k),
            # where n is number of rows and k is number of classes
            if len(preds.shape) == 3:
                preds = np.squeeze(preds, axis=1)

            grid.append("<OTHER>")
            ys.append(preds)
            counts.append(sum([row["count"] for row in feature_cnt[n_top_cats:]]))

        sample_df.unpersist()

        return grid, ys, counts

    @staticmethod
    def get_pdp_data_datetime_feature(
        df: SparkDataFrame,
        feature_name: str,
        model: Transformer,
        prediction_col: str,
        datetime_level: str,
        reader,
        ice_fraction: float = 1.0,
        ice_fraction_seed: int = 42,
    ) -> Tuple[List, List, List]:
        """Returns `grid`, `ys` and `counts` calculated on input datetime column to plot PDP.

        Args:
            df (SparkDataFrame): Spark DataFrame with `feature_name` column
            feature_name (str): feature column name
            model (PipelineModel): Spark Pipeline Model
            prediction_col (str): prediction column to be created by the `model`
            datetime_level (str): Unit of time that will be modified to calculate dependence: "year", "month" or "dayofweek"
            reader (_type_): Automl reader to transform input dataframe before `model` inferring.
            ice_fraction (float, optional): What fraction of the input dataframe will be used to make predictions.
                Useful for very large dataframe. Defaults to 1.0.
            ice_fraction_seed (int, optional): Seed for `ice_fraction`. Defaults to 42.

        Returns:
            Tuple[List, List, List]:
            `grid` is list of categories,
            `ys` is list of predictions by category,
            `counts` is numbers of values by category
        """
        df = reader.read(df).data
        if datetime_level == "year":
            feature_cnt = df.groupBy(sf.year(feature_name).alias("year")).count().orderBy(sf.asc("year")).collect()
            grid = [x["year"] for x in feature_cnt]
            counts = [row["count"] for row in feature_cnt]
            replace_date_element_udf = sf.udf(replace_year_in_date, DateType())
        elif datetime_level == "month":
            feature_cnt = df.groupBy(sf.month(feature_name).alias("month")).count().orderBy(sf.asc("month")).collect()
            grid = np.arange(1, 13)
            grid = grid.tolist()
            counts = [0] * 12
            for row in feature_cnt:
                counts[row["month"] - 1] = row["count"]
            replace_date_element_udf = sf.udf(replace_month_in_date, DateType())
        else:
            feature_cnt = (
                df.groupBy(sf.dayofweek(feature_name).alias("dayofweek")).count().orderBy(sf.asc("dayofweek")).collect()
            )
            grid = np.arange(7)
            grid = grid.tolist()
            counts = [0] * 7
            for row in feature_cnt:
                counts[row["dayofweek"] - 1] = row["count"]
            replace_date_element_udf = sf.udf(replace_dayofweek_in_date, DateType())
        ys = []
        sample_df = df.sample(fraction=ice_fraction, seed=ice_fraction_seed).cache()
        for i in tqdm(grid):
            feature_col = replace_date_element_udf(sf.col(feature_name), sf.lit(i)).alias(feature_name)
            sdf = sample_df.select(*[c for c in sample_df.columns if c != feature_name], feature_col)
            preds = model.transform(sdf)
            # TODO: SPARK-LAMA remove this line after passing the "prediction_col" parameter
            prediction_col = next(c for c in preds.columns if c.startswith("prediction"))
            preds = np.array(preds.select(prediction_col).collect())
            # when preds.shape is (n, 1, k) we change it to (n, k),
            # where n is number of rows and k is number of classes
            if len(preds.shape) == 3:
                preds = np.squeeze(preds, axis=1)
            ys.append(preds)

        return grid, ys, counts

    def get_individual_pdp(
        self,
        test_data: SparkDataFrame,
        feature_name: str,
        n_bins: Optional[int] = 30,
        top_n_categories: Optional[int] = 10,
        datetime_level: Optional[str] = "year",
        ice_fraction: float = 1.0,
        ice_fraction_seed: int = 42,
    ):
        assert feature_name in self.reader._roles
        assert datetime_level in ["year", "month", "dayofweek"]
        assert 0 < ice_fraction <= 1.0

        pipeline_model = self.transformer(leave_only_predict_cols=True)

        # Numerical features
        if self.reader._roles[feature_name].name == "Numeric":
            return self.get_pdp_data_numeric_feature(
                test_data, feature_name, pipeline_model, "prediction", n_bins, ice_fraction, ice_fraction_seed
            )
        # Categorical features
        elif self.reader._roles[feature_name].name == "Category":
            return self.get_pdp_data_categorical_feature(
                test_data, feature_name, pipeline_model, "prediction", top_n_categories, ice_fraction, ice_fraction_seed
            )
        # Datetime Features
        elif self.reader._roles[feature_name].name == "Datetime":
            return self.get_pdp_data_datetime_feature(
                test_data,
                feature_name,
                pipeline_model,
                "prediction",
                datetime_level,
                self.reader,
                ice_fraction,
                ice_fraction_seed,
            )
        else:
            raise ValueError("Supported only Numeric, Category or Datetime feature")

    def plot_pdp(
        self,
        test_data: ReadableToDf,
        feature_name: str,
        individual: Optional[bool] = False,
        n_bins: Optional[int] = 30,
        top_n_categories: Optional[int] = 10,
        top_n_classes: Optional[int] = 10,
        datetime_level: Optional[str] = "year",
        ice_fraction: float = 1.0,
        ice_fraction_seed: int = 42,
    ):
        grid, ys, counts = self.get_individual_pdp(
            test_data=test_data,
            feature_name=feature_name,
            n_bins=n_bins,
            top_n_categories=top_n_categories,
            datetime_level=datetime_level,
            ice_fraction=ice_fraction,
            ice_fraction_seed=ice_fraction_seed,
        )

        histogram_data_rows_limit = 2000
        rows_count = test_data.count()
        if rows_count > histogram_data_rows_limit:
            fraction = histogram_data_rows_limit / rows_count
            test_data = test_data.sample(frac=fraction)
        if self.reader._roles[feature_name].name == "Numeric":
            test_data = test_data.select(sf.col(feature_name).cast("double")).toPandas()
        else:
            test_data = test_data.select(feature_name).toPandas()
        plot_pdp_with_distribution(
            test_data,
            grid,
            ys,
            counts,
            self.reader,
            feature_name,
            individual,
            top_n_classes,
            datetime_level,
        )



import itertools

from collections import OrderedDict
from copy import deepcopy
from datetime import datetime
from typing import Dict
from typing import Iterator
from typing import List
from typing import Optional
from typing import Sequence
from typing import Set
from typing import cast

import holidays
import numpy as np
import pandas as pd

from lightautoml.dataset.base import RolesDict
from lightautoml.dataset.roles import CategoryRole
from lightautoml.dataset.roles import ColumnRole
from lightautoml.dataset.roles import DatetimeRole
from lightautoml.dataset.roles import NumericRole
from lightautoml.transformers.datetime import date_attrs
from lightautoml.transformers.datetime import datetime_check
from pyspark.ml import Transformer
from pyspark.ml.param.shared import Param
from pyspark.ml.param.shared import Params
from pyspark.sql import DataFrame as SparkDataFrame
from pyspark.sql import functions as sf
from pyspark.sql.pandas.functions import pandas_udf
from pyspark.sql.types import IntegerType

from sparklightautoml.mlwriters import CommonPickleMLReadable
from sparklightautoml.mlwriters import CommonPickleMLWritable
from sparklightautoml.transformers.base import SparkBaseEstimator
from sparklightautoml.transformers.base import SparkBaseTransformer
from sparklightautoml.transformers.scala_wrappers.is_holiday_transformer import (
    IsHolidayTransformer,
)


_DateSeasonsTransformations = Dict[str, List[str]]


def get_unit_of_timestamp_column(seas: str, col: str):
    """Generates pyspark column to extract unit of time from timestamp

    Args:
        seas (str): unit of time: `y`(year), `m`(month), `d`(day),
        `wd`(weekday), `hour`(hour), `min`(minute), `sec`(second), `ms`(microsecond), `ns`(nanosecond)
        col (str): column name with datetime values
    """
    if seas == "y":
        return sf.year(sf.to_timestamp(sf.col(col)))
    elif seas == "m":
        return sf.month(sf.to_timestamp(sf.col(col)))
    elif seas == "d":
        return sf.dayofmonth(sf.to_timestamp(sf.col(col)))
    # TODO SPARK-LAMA: F.dayofweek() starts numbering from another day.
    # Differs from pandas.Timestamp.weekday.
    # elif seas == 'wd':
    #     return F.dayofweek(F.to_timestamp(F.col(col)))
    elif seas == "hour":
        return sf.hour(sf.to_timestamp(sf.col(col)))
    elif seas == "min":
        return sf.minute(sf.to_timestamp(sf.col(col)))
    elif seas == "sec":
        return sf.second(sf.to_timestamp(sf.col(col)))
    else:

        @pandas_udf(IntegerType())
        def get_timestamp_attr(arrs: Iterator[pd.Series]) -> Iterator[pd.Series]:
            for x in arrs:

                def convert_to_datetime(timestamp: int):
                    try:
                        date = pd.to_datetime(datetime.fromtimestamp(timestamp))
                    except:
                        date = datetime.now()
                    return date

                x = x.apply(lambda d: convert_to_datetime(d))
                yield getattr(x.dt, date_attrs[seas])

        return get_timestamp_attr(sf.to_timestamp(sf.col(col)).cast("long"))


class SparkDatetimeHelper:
    """
    Helper class for :class:`~sparklightautoml.transformers.datetime.SparkTimeToNumTransformer`,
    :class:`~sparklightautoml.transformers.datetime.SparkBaseDiffTransformer` and
    :class:`~sparklightautoml.transformers.datetime.SparkDateSeasonsTransformer`
    """

    basic_interval = "D"

    _interval_mapping = {
        "NS": 0.000000001,
        "MS": 0.001,
        "SEC": 1,
        "MIN": 60,
        "HOUR": 60 * 60,
        "D": 60 * 60 * 24,
        # FIXME SPARK-LAMA: Very rough rounding
        "M": 60 * 60 * 24 * 30,
        "Y": 60 * 60 * 24 * 365,
    }

    _fit_checks = (datetime_check,)
    _transform_checks = ()


class SparkTimeToNumTransformer(
    SparkBaseTransformer, SparkDatetimeHelper, CommonPickleMLWritable, CommonPickleMLReadable
):
    """
    Transforms datetime columns values to numeric values.
    """

    basic_time = "2020-01-01"
    _fname_prefix = "dtdiff"

    def __init__(self, input_cols: List[str], input_roles: RolesDict, do_replace_columns: bool = False):
        output_cols = [f"{self._fname_prefix}__{f}" for f in input_cols]
        output_roles = {f: NumericRole(np.float32) for f in output_cols}
        super().__init__(input_cols, output_cols, input_roles, output_roles, do_replace_columns)

    def _transform(self, dataset: SparkDataFrame) -> SparkDataFrame:
        df = dataset

        new_cols = []
        for i, out_col in zip(self.getInputCols(), self.getOutputCols()):
            new_col = (
                (sf.to_timestamp(sf.col(i)).cast("long") - sf.to_timestamp(sf.lit(self.basic_time)).cast("long"))
                / self._interval_mapping[self.basic_interval]
            ).alias(out_col)
            new_cols.append(new_col)

        df = self._make_output_df(df, new_cols)

        return df


class SparkBaseDiffTransformer(
    SparkBaseTransformer, SparkDatetimeHelper, CommonPickleMLWritable, CommonPickleMLReadable
):
    """
    Basic conversion strategy, used in selection one-to-one transformers.
    Datetime converted to difference with basic_date.
    """

    _fname_prefix = "basediff"

    baseNames = Param(Params._dummy(), "baseNames", "base_names")

    diffNames = Param(Params._dummy(), "diffNames", "diff_names")

    basicInterval = Param(Params._dummy(), "basicInterval", "basic_interval")

    def __init__(
        self,
        input_roles: RolesDict,
        base_names: Sequence[str],
        diff_names: Sequence[str],
        basic_interval: Optional[str] = "D",
        do_replace_columns: bool = False,
    ):
        input_cols = list(base_names) + list(diff_names)

        self.base_names = base_names
        self.diff_names = diff_names
        self.basic_interval = basic_interval

        output_cols = [f"{self._fname_prefix}_{col}__{x}" for col in base_names for x in diff_names]

        output_roles = {col: NumericRole(dtype=np.float32) for col in output_cols}

        super().__init__(input_cols, output_cols, input_roles, output_roles, do_replace_columns)

        self.set(self.baseNames, self.base_names)
        self.set(self.diffNames, self.diff_names)
        self.set(self.basicInterval, self.basic_interval)

    def _transform(self, df: SparkDataFrame) -> SparkDataFrame:
        new_cols = [
            (
                (sf.to_timestamp(sf.col(dif)).cast("long") - sf.to_timestamp(sf.col(base)).cast("long"))
                / self._interval_mapping[self.basic_interval]
            ).alias(f"{self._fname_prefix}_{base}__{dif}")
            for base in self.base_names
            for dif in self.diff_names
        ]

        df = self._make_output_df(df, new_cols)

        return df


class SparkDateSeasonsEstimator(SparkBaseEstimator):
    _fname_prefix = "season"

    def __init__(self, input_cols: List[str], input_roles: RolesDict, output_role: Optional[ColumnRole] = None):
        output_role = output_role or CategoryRole(np.int32)

        super().__init__(input_cols, input_roles, output_role=output_role)
        self._infer_output_cols_and_roles(output_role)

    def _infer_output_cols_and_roles(self, output_role: ColumnRole):
        input_roles = self.get_input_roles()
        self.transformations: _DateSeasonsTransformations = OrderedDict()
        self.seasons_out_cols: Dict[str, List[str]] = dict()
        self.holidays_out_cols = []
        for col in self.getInputCols():
            rdt = cast(DatetimeRole, input_roles[col])
            seas = list(rdt.seasonality)
            self.transformations[col] = seas
            self.seasons_out_cols[col] = [f"{self._fname_prefix}_{s}__{col}" for s in seas]
            if rdt.country is not None:
                self.holidays_out_cols.append(f"{self._fname_prefix}_hol__{col}")

        output_cols = [*itertools.chain(*self.seasons_out_cols.values()), *self.holidays_out_cols]
        output_roles = {f: deepcopy(output_role) for f in output_cols}

        self.set(self.outputCols, output_cols)
        self.set(self.outputRoles, output_roles)

    def _fit(self, dataset: SparkDataFrame) -> Transformer:
        roles = self.get_input_roles()

        min_max_years = (
            dataset.select(
                *[
                    sf.struct(sf.year(sf.min(in_col)).alias("min"), sf.year(sf.max(in_col)).alias("max")).alias(in_col)
                    for in_col in self.getInputCols()
                ]
            )
            .first()
            .asDict()
        )

        holidays_cols_dates: Dict[str, Set[str]] = {
            col: set(
                dt.strftime("%Y-%m-%d")
                for dt in holidays.country_holidays(
                    years=list(range(min_y, max_y + 1)),
                    country=roles[col].country,
                    prov=roles[col].prov,
                    state=roles[col].state,
                ).keys()
            )
            for col, (min_y, max_y) in min_max_years.items()
        }

        return SparkDateSeasonsTransformer(
            input_cols=self.getInputCols(),
            seasons_out_cols=self.seasons_out_cols,
            holidays_out_cols=self.holidays_out_cols,
            input_roles=self.get_input_roles(),
            output_roles=self.get_output_roles(),
            seasons_transformations=self.transformations,
            holidays_dates=holidays_cols_dates,
        )


class SparkDateSeasonsTransformer(
    SparkBaseTransformer, SparkDatetimeHelper, CommonPickleMLWritable, CommonPickleMLReadable
):
    """
    Extracts unit of time from Datetime values and marks holiday dates.
    """

    _fname_prefix = "season"

    seasonOutCols = Param(Params._dummy(), "seasonOutCols", "seasonOutCols")
    holidaysOutCols = Param(Params._dummy(), "holidaysOutCols", "holidaysOutCols")
    seasonsTransformations = Param(Params._dummy(), "seasonsTransformations", "seasonsTransformations")
    holidaysDates = Param(Params._dummy(), "holidaysDates", "holidaysDates")

    def __init__(
        self,
        input_cols: List[str],
        seasons_out_cols: Dict[str, List[str]],
        holidays_out_cols: List[str],
        input_roles: RolesDict,
        output_roles: RolesDict,
        seasons_transformations: _DateSeasonsTransformations,
        holidays_dates: Dict[str, Set[str]],
    ):
        output_cols = [*itertools.chain(*seasons_out_cols.values()), *holidays_out_cols]
        super().__init__(input_cols, output_cols, input_roles, output_roles)

        self.set(self.seasonOutCols, seasons_out_cols)
        self.set(self.holidaysOutCols, holidays_out_cols)
        self.set(self.seasonsTransformations, seasons_transformations)
        self.set(self.holidaysDates, holidays_dates)

    def _transform(self, dataset: SparkDataFrame) -> SparkDataFrame:
        df = dataset

        seasons_out_cols = self.getOrDefault(self.seasonOutCols)
        holidays_out_cols = self.getOrDefault(self.holidaysOutCols)
        seasons_transformations = self.getOrDefault(self.seasonsTransformations)
        holidays_dates = self.getOrDefault(self.holidaysDates)

        new_cols = []
        for col, transformations in seasons_transformations.items():
            fcol = sf.to_timestamp(sf.col(col)).cast("long")
            seas_cols = [
                (
                    sf.when(sf.isnan(fcol) | sf.isnull(fcol), None)
                    .otherwise(get_unit_of_timestamp_column(seas, col))
                    .alias(out_col)
                )
                for out_col, seas in zip(seasons_out_cols[col], transformations)
            ]

            new_cols.extend(seas_cols)

        holidays_transformer = IsHolidayTransformer.create(
            holidays_dates=holidays_dates, input_cols=self.getInputCols(), output_cols=holidays_out_cols
        )
        df = holidays_transformer.transform(df.select("*", *new_cols))

        return df
